# -*- coding:utf-8 -*-
import unittest

from tyme4py.solar import SolarSeason


class TestSolatSeason(unittest.TestCase):
    def test0(self):
        season: SolarSeason = SolarSeason(2023, 0)
        assert season.__str__() == '2023年一季度'
        assert season.next(-5).__str__() == '2021年四季度'
        assert season.next(5).__str__() == '2024年二季度'



# -*- coding:utf-8 -*-
import unittest

from tyme4py.culture import Element
from tyme4py.sixtycycle import HeavenStem, EarthBranch


class TestElement(unittest.TestCase):
    def test(self):
        assert Element('金').get_restrain().get_name() == Element('木').get_name()

    def test1(self):
        assert Element('火').get_reinforce().get_name() == Element('土').get_name()

    def test2(self):
        assert HeavenStem('丙').get_element().get_name() == '火'

    def test3(self):
        assert EarthBranch('寅').get_element().get_name() == '木'
        assert EarthBranch('寅').get_element().get_reinforce().get_name() == Element('火').get_name()



# -*- coding:utf-8 -*-
from __future__ import annotations

from typing import List, Union

from tyme4py import LoopTyme, AbstractCultureDay
from tyme4py.jd import JulianDay
from tyme4py.util import ShouXingUtil


class Phenology(LoopTyme):
    """候"""
    NAMES: List[str] = ['蚯蚓结', '麋角解', '水泉动', '雁北乡', '鹊始巢', '雉始雊', '鸡始乳', '征鸟厉疾', '水泽腹坚', '东风解冻', '蛰虫始振', '鱼陟负冰', '獭祭鱼', '候雁北', '草木萌动', '桃始华', '仓庚鸣', '鹰化为鸠', '玄鸟至', '雷乃发声', '始电', '桐始华', '田鼠化为鴽', '虹始见', '萍始生', '鸣鸠拂其羽', '戴胜降于桑', '蝼蝈鸣', '蚯蚓出', '王瓜生', '苦菜秀', '靡草死', '麦秋至', '螳螂生', '鵙始鸣', '反舌无声', '鹿角解', '蜩始鸣', '半夏生', '温风至', '蟋蟀居壁', '鹰始挚', '腐草为萤', '土润溽暑', '大雨行时', '凉风至', '白露降', '寒蝉鸣', '鹰乃祭鸟', '天地始肃', '禾乃登', '鸿雁来', '玄鸟归', '群鸟养羞', '雷始收声', '蛰虫坯户', '水始涸', '鸿雁来宾', '雀入大水为蛤', '菊有黄花', '豺乃祭兽', '草木黄落', '蛰虫咸俯', '水始冰', '地始冻', '雉入大水为蜃', '虹藏不见', '天气上升地气下降', '闭塞而成冬', '鹖鴠不鸣', '虎始交', '荔挺出']
    """名称"""

    _year: int
    """年"""

    def __init__(self, year: int, index_or_name: Union[int, str]):
        super().__init__(self.NAMES, index_or_name)
        if isinstance(index_or_name, int):
            size: int = self.get_size()
            self._year = (year * size + index_or_name) // size
        else:
            self._year = year

    @classmethod
    def from_name(cls, year: int, name: str) -> Phenology:
        return cls(year, name)

    @classmethod
    def from_index(cls, year: int, index: int) -> Phenology:
        return cls(year, index)

    def next(self, n: int) -> Phenology:
        size: int = self.get_size()
        i: int = self.get_index() + n
        return Phenology((self.get_year() * size + i) // size, self.index_of(i, size))

    def get_year(self) -> int:
        """
        :return: 年
        """
        return self._year

    def get_three_phenology(self) -> ThreePhenology:
        """
        :return: 三候
        """
        return ThreePhenology(self.get_index() % 3)

    def get_julian_day(self) -> JulianDay:
        """
        :return: 儒略日
        """
        t = ShouXingUtil.sa_lon_t((self.get_year() - 2000 + (self.get_index() - 18) * 5.0 / 360 + 1) * 2 * ShouXingUtil.PI)
        return JulianDay.from_julian_day(t * 36525 + JulianDay.J2000 + 8.0 / 24 - ShouXingUtil.dt_t(t * 36525))


class ThreePhenology(LoopTyme):
    """三候"""
    NAMES: List[str] = ['初候', '二候', '三候']
    """名称"""

    def __init__(self, index_or_name: Union[int, str]):
        super().__init__(self.NAMES, index_or_name)

    @classmethod
    def from_name(cls, name: str) -> ThreePhenology:
        return cls(name)

    @classmethod
    def from_index(cls, index: int) -> ThreePhenology:
        return cls(index)

    def next(self, n: int) -> ThreePhenology:
        return ThreePhenology(self.next_index(n))

    def get_three_phenology(self) -> ThreePhenology:
        return ThreePhenology(self.get_index() % 3)


class PhenologyDay(AbstractCultureDay):
    """七十二候"""

    def __init__(self, phenology: Phenology, day_index: int):
        super().__init__(phenology, day_index)

    def get_phenology(self) -> Phenology:
        return super().get_culture()



# -*- coding:utf-8 -*-
import unittest

from tyme4py.solar import SolarYear


class TestSolarYear(unittest.TestCase):
    def test0(self):
        assert SolarYear(2023).get_name() == '2023年'

    def test1(self):
        assert SolarYear(2023).is_leap() is False

    def test2(self):
        assert SolarYear(1500).is_leap() is True

    def test3(self):
        assert SolarYear(1700).is_leap() is False

    def test4(self):
        assert SolarYear(2023).get_day_count() == 365

    def test5(self):
        assert SolarYear(2023).next(5).get_name() == '2028年'

    def test6(self):
        assert SolarYear(2023).next(-5).get_name() == '2018年'



# -*- coding:utf-8 -*-
import unittest

from tyme4py.culture import Zodiac
from tyme4py.rabbyung import RabByungDay, RabByungElement
from tyme4py.solar import SolarDay


class TestRabByungDay(unittest.TestCase):
    def test0(self):
        # 测试公历1951年1月8日转换为藏历日
        solar_day = SolarDay.from_ymd(1951, 1, 8)
        rab_day = RabByungDay.from_solar_day(solar_day)
        self.assertEqual("第十六饶迥铁虎年十二月初一", str(rab_day))

        # 测试藏历日转换为公历日
        rab_day2 = RabByungDay.from_element_zodiac(15, RabByungElement.from_name("铁"), Zodiac.from_name("虎"), 12, 1)
        self.assertEqual("1951年1月8日", str(rab_day2.get_solar_day()))

    def test1(self):
        # 测试公历2051年2月11日转换为藏历日
        solar_day = SolarDay.from_ymd(2051, 2, 11)
        rab_day = RabByungDay.from_solar_day(solar_day)
        self.assertEqual("第十八饶迥铁马年十二月三十", str(rab_day))

        # 测试藏历日转换为公历日
        rab_day2 = RabByungDay.from_element_zodiac(17, RabByungElement.from_name("铁"), Zodiac.from_name("马"), 12, 30)
        self.assertEqual("2051年2月11日", str(rab_day2.get_solar_day()))

    def test2(self):
        # 测试公历2025年4月23日转换为藏历日
        solar_day = SolarDay.from_ymd(2025, 4, 23)
        rab_day = RabByungDay.from_solar_day(solar_day)
        self.assertEqual("第十七饶迥木蛇年二月廿五", str(rab_day))

        # 测试藏历日转换为公历日
        rab_day2 = RabByungDay.from_element_zodiac(16, RabByungElement.from_name("木"), Zodiac.from_name("蛇"), 2, 25)
        self.assertEqual("2025年4月23日", str(rab_day2.get_solar_day()))

    def test3(self):
        # 测试公历1951年2月8日转换为藏历日
        solar_day = SolarDay.from_ymd(1951, 2, 8)
        rab_day = RabByungDay.from_solar_day(solar_day)
        self.assertEqual("第十六饶迥铁兔年正月初二", str(rab_day))

        # 测试藏历日转换为公历日
        rab_day2 = RabByungDay.from_element_zodiac(15, RabByungElement.from_name("铁"), Zodiac.from_name("兔"), 1, 2)
        self.assertEqual("1951年2月8日", str(rab_day2.get_solar_day()))

    def test4(self):
        # 测试公历1951年1月24日转换为藏历日（闰日）
        solar_day = SolarDay.from_ymd(1951, 1, 24)
        rab_day = RabByungDay.from_solar_day(solar_day)
        self.assertEqual("第十六饶迥铁虎年十二月闰十六", str(rab_day))

        # 测试藏历日转换为公历日（闰日）
        rab_day2 = RabByungDay.from_element_zodiac(15, RabByungElement.from_name("铁"), Zodiac.from_name("虎"), 12, -16)
        self.assertEqual("1951年1月24日", str(rab_day2.get_solar_day()))

    def test5(self):
        # 测试公历1961年6月24日转换为藏历日
        solar_day = SolarDay.from_ymd(1961, 6, 24)
        rab_day = RabByungDay.from_solar_day(solar_day)
        self.assertEqual("第十六饶迥铁牛年五月十一", str(rab_day))

        # 测试藏历日转换为公历日
        rab_day2 = RabByungDay.from_element_zodiac(15, RabByungElement.from_name("铁"), Zodiac.from_name("牛"), 5, 11)
        self.assertEqual("1961年6月24日", str(rab_day2.get_solar_day()))

    def test6(self):
        # 测试公历1952年2月23日转换为藏历日
        solar_day = SolarDay.from_ymd(1952, 2, 23)
        rab_day = RabByungDay.from_solar_day(solar_day)
        self.assertEqual("第十六饶迥铁兔年十二月廿八", str(rab_day))

        # 测试藏历日转换为公历日
        rab_day2 = RabByungDay.from_element_zodiac(15, RabByungElement.from_name("铁"), Zodiac.from_name("兔"), 12, 28)
        self.assertEqual("1952年2月23日", str(rab_day2.get_solar_day()))

    def test7(self):
        # 测试公历2025年4月26日转换为藏历日
        solar_day = SolarDay.from_ymd(2025, 4, 26)
        rab_day = RabByungDay.from_solar_day(solar_day)
        self.assertEqual("第十七饶迥木蛇年二月廿九", str(rab_day))

    def test8(self):
        # 测试公历2025年4月25日转换为藏历日
        solar_day = SolarDay.from_ymd(2025, 4, 25)
        rab_day = RabByungDay.from_solar_day(solar_day)
        self.assertEqual("第十七饶迥木蛇年二月廿七", str(rab_day))



# -*- coding:utf-8 -*-
from __future__ import annotations

from typing import List, Union

from tyme4py import LoopTyme
from tyme4py.culture import Week


class SevenStar(LoopTyme):
    """七曜（七政、七纬、七耀）"""
    NAMES: List[str] = ['日', '月', '火', '水', '木', '金', '土']
    """名称"""

    def __init__(self, index_or_name: Union[int, str]):
        super().__init__(self.NAMES, index_or_name)

    @classmethod
    def from_name(cls, name: str) -> SevenStar:
        return cls(name)

    @classmethod
    def from_index(cls, index: int) -> SevenStar:
        return cls(index)

    def next(self, n: int) -> SevenStar:
        return SevenStar(super().next_index(n))

    def get_week(self) -> Week:
        """
        :return:星期
        """
        return Week(self.get_index())



# -*- coding:utf-8 -*-
from __future__ import annotations

from typing import List, Union

from tyme4py import LoopTyme, AbstractCultureDay


class Dog(LoopTyme):
    """三伏"""
    NAMES: List[str] = ['初伏', '中伏', '末伏']
    """名称"""

    def __init__(self, index_or_name: Union[int, str]):
        super().__init__(self.NAMES, index_or_name)

    @classmethod
    def from_name(cls, name: str) -> Dog:
        return cls(name)

    @classmethod
    def from_index(cls, index: int) -> Dog:
        return cls(index)

    def next(self, n: int) -> Dog:
        return Dog(self.next_index(n))


class DogDay(AbstractCultureDay):
    """三伏天"""

    def __init__(self, dog: Dog, day_index: int):
        super().__init__(dog, day_index)

    def get_dog(self) -> Dog:
        """
        三伏
        :return: 三伏
        """
        return self.get_culture()



# -*- coding:utf-8 -*-
import unittest

from tyme4py.solar import SolarDay, SolarWeek


class TestWeek(unittest.TestCase):
    def test0(self):
        assert SolarDay(1582, 10, 1).get_week().get_name() == '一'

    def test1(self):
        assert SolarDay(1582, 10, 15).get_week().get_name() == '五'

    def test2(self):
        assert SolarDay(2023, 10, 31).get_week().get_index() == 2

    def test3(self):
        w = SolarWeek(2023, 10, 0, 0)
        assert w.get_name() == '第一周'
        assert w.__str__() == '2023年10月第一周'

    def test5(self):
        w = SolarWeek(2023, 10, 4, 0)
        assert w.get_name() == '第五周'
        assert w.__str__() == '2023年10月第五周'

    def test6(self):
        w = SolarWeek(2023, 10, 5, 1)
        assert w.get_name() == '第六周'
        assert w.__str__() == '2023年10月第六周'

    def test7(self):
        w = SolarWeek(2023, 10, 0, 0).next(4)
        assert w.get_name() == '第五周'
        assert w.__str__() == '2023年10月第五周'

    def test8(self):
        w = SolarWeek(2023, 10, 0, 0).next(5)
        assert w.get_name() == '第二周'
        assert w.__str__() == '2023年11月第二周'

    def test9(self):
        w = SolarWeek(2023, 10, 0, 0).next(-1)
        assert w.get_name() == '第五周'
        assert w.__str__() == '2023年9月第五周'

    def test10(self):
        w = SolarWeek(2023, 10, 0, 0).next(-5)
        assert w.get_name() == '第一周'
        assert w.__str__() == '2023年9月第一周'

    def test11(self):
        w = SolarWeek(2023, 10, 0, 0).next(-6)
        assert w.get_name() == '第四周'
        assert w.__str__() == '2023年8月第四周'

    def test12(self):
        assert SolarDay(1582, 10, 1).get_week().get_index() == 1

    def test13(self):
        assert SolarDay(1582, 10, 15).get_week().get_index() == 5

    def test14(self):
        assert SolarDay(1129, 11, 17).get_week().get_index() == 0

    def test15(self):
        assert SolarDay(1129, 11, 1).get_week().get_index() == 5

    def test16(self):
        assert SolarDay(8, 11, 1).get_week().get_index() == 4

    def test17(self):
        assert SolarDay(1582, 9, 30).get_week().get_index() == 0

    def test18(self):
        assert SolarDay(1582, 1, 1).get_week().get_index() == 1

    def test19(self):
        assert SolarDay(1500, 2, 29).get_week().get_index() == 6

    def test20(self):
        assert SolarDay(9865, 7, 26).get_week().get_index() == 3

    def test21(self):
        start = 0
        week: SolarWeek = SolarWeek(2024, 2, 2, start)
        assert week.__str__() == '2024年2月第三周'
        assert week.get_index_in_year() == 6
        week = SolarDay(2024, 2, 11).get_solar_week(start)
        assert week.__str__() == '2024年2月第三周'
        week = SolarDay(2024, 2, 17).get_solar_week(start)
        assert week.__str__() == '2024年2月第三周'
        week = SolarDay(2024, 2, 10).get_solar_week(start)
        assert week.__str__() == '2024年2月第二周'
        week = SolarDay(2024, 2, 18).get_solar_week(start)
        assert week.__str__() == '2024年2月第四周'

    def test22(self):
        start: int = 0
        week: SolarWeek = SolarDay(2024, 7, 1).get_solar_week(start)
        assert week.__str__(), '2024年7月第一周'
        assert week.get_index_in_year() == 26
        week = week.next(1)
        assert week.__str__() == '2024年7月第二周'



# -*- coding:utf-8 -*-
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from tyme4py.eightchar import EightChar
    from tyme4py.lunar import LunarHour
    from tyme4py.solar import SolarTime, SolarTerm
    from tyme4py.eightchar import ChildLimitInfo


class EightCharProvider(ABC):
    """八字计算接口"""

    @abstractmethod
    def get_eight_char(self, hour: LunarHour) -> EightChar:
        """
        八字
        :param hour: 农历时辰
        :return: 八字 EightChar
        """
        pass


class ChildLimitProvider(ABC):
    """童限计算接口"""

    @abstractmethod
    def get_info(self, birth_time: SolarTime, term: SolarTerm) -> ChildLimitInfo:
        """
        童限信息
        :param birth_time:出生公历时刻
        :param term: 节令
        :return: 童限信息
        """
        pass



# -*- coding:utf-8 -*-
import unittest

from tyme4py.sixtycycle import EarthBranch


class TestEarthBranch(unittest.TestCase):
    def test(self):
        assert EarthBranch(0).get_name() == '子'

    def test1(self):
        assert EarthBranch(0).get_index() == 0

    def test2(self):
        assert EarthBranch('子').get_opposite().get_name() == '午'
        assert EarthBranch('戌').get_opposite().get_name() == '辰'

    def test3(self):
        assert EarthBranch('子').get_combine().get_name() == '丑'
        assert EarthBranch('申').get_combine().get_name() == '巳'

    def test4(self):
        assert EarthBranch('巳').get_harm().get_name() == '寅'
        assert EarthBranch('申').get_harm().get_name() == '亥'

    def test5(self):
        # 合化
        assert EarthBranch('卯').combine(EarthBranch('戌')).get_name() == '火'
        assert EarthBranch('戌').combine(EarthBranch('卯')).get_name() == '火'
        # 卯子无法合化
        assert EarthBranch('卯').combine(EarthBranch('子')) is None



# -*- coding:utf-8 -*-
import unittest

from tyme4py.rabbyung import RabByungMonth


class TestRabByungMonth(unittest.TestCase):
    def test0(self):
        assert "第十六饶迥铁虎年十二月" == RabByungMonth.from_ym(1950, 12).__str__()



# -*- coding:utf-8 -*-
from __future__ import annotations

from typing import Union, List

from tyme4py import LoopTyme, AbstractCultureDay


class Nine(LoopTyme):
    """数九"""
    NAMES: List[str] = ['一九', '二九', '三九', '四九', '五九', '六九', '七九', '八九', '九九']
    """名称"""

    def __init__(self, index_or_name: Union[int, str]):
        super().__init__(self.NAMES, index_or_name)

    @classmethod
    def from_name(cls, name: str) -> Nine:
        return cls(name)

    @classmethod
    def from_index(cls, index: int) -> Nine:
        return cls(index)

    def next(self, n: int):
        return Nine(self.next_index(n))


class NineDay(AbstractCultureDay):
    """数九天"""

    def __init__(self, nine: Nine, day_index: int):
        super().__init__(nine, day_index)

    def get_nine(self) -> Nine:
        """
        数九
        :return: 数九
        """
        return self.get_culture()




# -*- coding:utf-8 -*-
from __future__ import annotations

from math import floor, ceil
from typing import TYPE_CHECKING, Union, List

from tyme4py import LoopTyme, AbstractCultureDay
from tyme4py.culture import Week, Constellation, PhaseDay, Phase
from tyme4py.culture.dog import DogDay, Dog
from tyme4py.culture.nine import NineDay, Nine
from tyme4py.culture.phenology import PhenologyDay, Phenology
from tyme4py.culture.plumrain import PlumRainDay, PlumRain
from tyme4py.enums import HideHeavenStemType
from tyme4py.unit import YearUnit, MonthUnit, WeekUnit, DayUnit, SecondUnit
from tyme4py.util import ShouXingUtil
from tyme4py.jd import JulianDay
from tyme4py.festival import SolarFestival
from tyme4py.holiday import LegalHoliday

if TYPE_CHECKING:
    from tyme4py.lunar import LunarDay, LunarMonth, LunarHour
    from tyme4py.sixtycycle import HideHeavenStemDay, SixtyCycleDay, SixtyCycleHour
    from tyme4py.rabbyung import RabByungYear, RabByungDay


class SolarTerm(LoopTyme):
    """节气"""
    NAMES = ['冬至', '小寒', '大寒', '立春', '雨水', '惊蛰', '春分', '清明', '谷雨', '立夏', '小满', '芒种', '夏至', '小暑', '大暑', '立秋', '处暑', '白露', '秋分', '寒露', '霜降', '立冬', '小雪', '大雪']
    """名称"""
    _year: int
    """年"""
    _cursory_julian_day: float
    """儒略日（用于日历，只精确到日中午12:00）"""

    def __init__(self, year: int, index_or_name: Union[int, str]):
        super().__init__(self.NAMES, index_or_name)
        y = year
        if isinstance(index_or_name, int):
            size = self.NAMES.__len__()
            y = (year * size + index_or_name) // size
        jd = floor((y - 2000) * 365.2422 + 180)
        # 355对应2000.12冬至，计算粗略估计值
        w = floor((jd - 355 + 183) / 365.2422) * 365.2422 + 355
        if ShouXingUtil.calc_qi(w) > jd:
            w -= 365.2422
        self._year = y
        self._cursory_julian_day = ShouXingUtil.calc_qi(w + 15.2184 * self.get_index())

    @classmethod
    def from_index(cls, year: int, index: int) -> SolarTerm:
        return cls(year, index)

    @classmethod
    def from_name(cls, year: int, name: str) -> SolarTerm:
        return cls(year, name)

    def next(self, n: int) -> SolarTerm:
        size = self.get_size()
        i = self.get_index() + n
        return SolarTerm((self._year * size + i) // size, self._index_of_by(i))

    def is_jie(self) -> bool:
        """是否节令"""
        return self._index % 2 == 1

    def is_qi(self) -> bool:
        """是否气令"""
        return self._index % 2 == 0

    def get_julian_day(self) -> JulianDay:
        """
        儒略日（精确到秒）
        :return : 儒略日
        """
        return JulianDay.from_julian_day(ShouXingUtil.qi_accurate2(self._cursory_julian_day) + JulianDay.J2000)

    def get_solar_day(self) -> SolarDay:
        """
        公历日（用于日历）
        @return: 公历日
        """
        return JulianDay.from_julian_day(self._cursory_julian_day + JulianDay.J2000).get_solar_day()

    def get_year(self) -> int:
        return self._year

    def get_cursory_julian_day(self) -> float:
        """
        儒略日（用于日历，只精确到日中午12:00）
        @return: 儒略日
        """
        return self._cursory_julian_day


class SolarTermDay(AbstractCultureDay):
    """节气第几天"""

    def __init__(self, solar_term: SolarTerm, day_index: int):
        super().__init__(solar_term, day_index)

    def get_solar_term(self) -> SolarTerm:
        """
        :return: 节气
        """
        return self.get_culture()


class SolarYear(YearUnit):
    """公历年"""

    def __init__(self, year: int):
        SolarYear.validate(year)
        super().__init__(year)

    @staticmethod
    def validate(year: int) -> None:
        if year < 1 or year > 9999:
            raise ValueError(f'illegal solar year: {year}')

    @classmethod
    def from_year(cls, year: int) -> SolarYear:
        return cls(year)

    def get_day_count(self) -> int:
        if self._year == 1582:
            return 355
        return 366 if self.is_leap() else 365

    def is_leap(self) -> bool:
        if self._year < 1600:
            return self._year % 4 == 0
        return (self._year % 4 == 0 and self._year % 100 != 0) or (self._year % 400 == 0)

    def get_name(self) -> str:
        return f"{self._year}年"

    def next(self, n: int) -> SolarYear:
        return SolarYear.from_year(self._year + n)

    def get_months(self) -> List[SolarMonth]:
        return [SolarMonth.from_ym(self._year, month) for month in range(1, 13)]

    def get_seasons(self) -> List[SolarSeason]:
        return [SolarSeason.from_index(self._year, i) for i in range(4)]

    def get_half_years(self) -> List[SolarHalfYear]:
        return [SolarHalfYear.from_index(self._year, i) for i in range(2)]

    def get_rab_byung_year(self) -> RabByungYear:
        from tyme4py.rabbyung import RabByungYear
        return RabByungYear.from_year(self._year)


class SolarHalfYear(YearUnit):
    """公历半年"""
    NAMES: List[str] = ['上半年', '下半年']
    _index: int
    """索引"""

    @staticmethod
    def validate(year: int, index: int) -> None:
        if index < 0 or index > 1:
            raise ValueError(f'illegal solar half year index: {index}')
        SolarYear.validate(year)

    def __init__(self, year: int, index: int):
        SolarHalfYear.validate(year, index)
        super().__init__(year)
        self._index = index

    @classmethod
    def from_index(cls, year: int, index: int) -> SolarHalfYear:
        return cls(year, index)

    def get_solar_year(self) -> SolarYear:
        return SolarYear.from_year(self._year)

    def get_index(self) -> int:
        return self._index

    def get_name(self) -> str:
        return self.NAMES[self._index]

    def __str__(self) -> str:
        return f'{self.get_solar_year()}{self.get_name()}'

    def next(self, n: int) -> SolarHalfYear:
        i = self._index + n
        return SolarHalfYear.from_index((self.get_year() * 2 + i) // 2, self.index_of(i, 2))

    def get_months(self) -> List[SolarMonth]:
        y = self.get_year()
        return [SolarMonth.from_ym(y, self._index * 6 + i) for i in range(1, 7)]

    def get_seasons(self) -> List[SolarSeason]:
        y = self.get_year()
        return [SolarSeason.from_index(y, self._index * 2 + i) for i in range(2)]


class SolarSeason(YearUnit):
    """公历季度"""
    NAMES = ['一季度', '二季度', '三季度', '四季度']

    @staticmethod
    def validate(year: int, index: int) -> None:
        if index < 0 or index > 3:
            raise ValueError(f'illegal solar season index: {index}')
        SolarYear.validate(year)

    def __init__(self, year: int, index: int):
        SolarSeason.validate(year, index)
        super().__init__(year)
        self._index = index

    @classmethod
    def from_index(cls, year: int, index: int) -> SolarSeason:
        return cls(year, index)

    def get_solar_year(self) -> SolarYear:
        return SolarYear.from_year(self._year)

    def get_index(self) -> int:
        return self._index

    def get_name(self) -> str:
        return self.NAMES[self._index]

    def __str__(self) -> str:
        return f'{self.get_solar_year()}{self.get_name()}'

    def next(self, n: int) -> SolarSeason:
        i = self._index + n
        return SolarSeason.from_index((self.get_year() * 4 + i) // 4, self.index_of(i, 4))

    def get_months(self) -> List[SolarMonth]:
        y = self.get_year()
        return [SolarMonth.from_ym(y, self._index * 3 + i) for i in range(1, 4)]


class SolarMonth(MonthUnit):
    """公历月"""
    NAMES: List[str] = ['1月', '2月', '3月', '4月', '5月', '6月', '7月', '8月', '9月', '10月', '11月', '12月']
    _DAYS: List[int] = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
    """每月天数"""

    @staticmethod
    def validate(year: int, month: int) -> None:
        if month < 1 or month > 12:
            raise ValueError(f'illegal solar month: {month}')
        SolarYear.validate(year)

    def __init__(self, year: int, month: int):
        """
        :param year:  年
        :param month: 月
        """
        SolarMonth.validate(year, month)
        super().__init__(year, month)

    @classmethod
    def from_ym(cls, year: int, month: int) -> SolarMonth:
        return cls(year, month)

    def get_solar_year(self) -> SolarYear:
        """
        :return: 公历年
        """
        return SolarYear.from_year(self._year)

    def get_day_count(self) -> int:
        """
        当月的总天数
        :return:  天数（1582年10月只有21天)
        """
        if 1582 == self._year and 10 == self._month:
            return 21
        d: int = self._DAYS[self.get_index_in_year()]
        # 公历闰年2月多一天
        if 2 == self._month and self.get_solar_year().is_leap():
            d += 1
        return d

    def get_index_in_year(self) -> int:
        """
        位于当年的月索引
        :return: 位于当年的索引(0-11)
        """
        return self._month - 1

    def get_season(self) -> SolarSeason:
        """
        :return: 公历季度  SolarSeason
        """
        return SolarSeason(self._year, int(self.get_index_in_year() / 3))

    def get_week_count(self, start: int) -> int:
        """
        当月有几周
        :param start:起始星期，1234560分别代表星期一至星期天
        :return:周数
        """
        return ceil((self.index_of(SolarDay(self._year, self._month, 1).get_week().get_index() - start, 7) + self.get_day_count()) / 7)

    def get_name(self) -> str:
        return self.NAMES[self.get_index_in_year()]

    def __str__(self) -> str:
        return f'{self.get_solar_year()}{self.get_name()}'

    def next(self, n: int) -> SolarMonth:
        i = self._month - 1 + n
        return SolarMonth.from_ym((self._year * 12 + i) // 12, self.index_of(i, 12) + 1)

    def get_weeks(self, start: int) -> List[SolarWeek]:
        """
        本月的公历周列表
        :param start: 星期几作为一周的开始，1234560分别代表星期一至星期天
        :return: 周列表
        """
        l: List[SolarWeek] = []
        for i in range(0, self.get_week_count(start)):
            l.append(SolarWeek(self._year, self._month, i, start))
        return l

    def get_days(self) -> List[SolarDay]:
        """
        本月的公历日列表
        :return: 公历日 SolarDay 的列表，从1日开始
        """
        l: List[SolarDay] = []
        for i in range(1, self.get_day_count()):
            l.append(SolarDay(self._year, self._month, i))
        return l

    def get_first_day(self) -> SolarDay:
        """
        本月第一天
        :return: 公历日 SolarDay
        """
        return SolarDay.from_ymd(self._year, self._month, 1)


class SolarWeek(WeekUnit):
    """公历周"""
    NAMES: List[str] = ['第一周', '第二周', '第三周', '第四周', '第五周', '第六周']

    @staticmethod
    def validate(year: int, month: int, index: int, start: int) -> None:
        WeekUnit.validate(year, month, index, start)
        m: SolarMonth = SolarMonth(year, month)
        if index >= m.get_week_count(start):
            raise ValueError(f'illegal solar week index: {index} in month: {m}')

    def __init__(self, year: int, month: int, index: int, start: int):
        """
        :param year:  年
        :param month: 月
        :param index: 索引，0-5
        :param start: 起始星期，1234560分别代表星期一至星期天
        """
        SolarWeek.validate(year, month, index, start)
        super().__init__(year, month, index, start)

    @classmethod
    def from_ym(cls, year: int, month: int, index: int, start: int) -> SolarWeek:
        return cls(year, month, index, start)

    def get_solar_month(self) -> SolarMonth:
        """
        :return: 公历月
        """
        return SolarMonth.from_ym(self._year, self._month)

    def get_index_in_year(self) -> int:
        """
        位于当年的索引
        :return: 索引  注意：索引值是从0开始，即0代表第一周
        """
        i: int = 0
        # 本周第1天
        first_day: SolarDay = self.get_first_day()
        # 今年第1周
        w: SolarWeek = SolarWeek(self._year, 1, 0, self._start)
        while not w.get_first_day() == first_day:
            w = w.next(1)
            i += 1
        return i

    def get_name(self) -> str:
        return self.NAMES[self._index]

    def __str__(self) -> str:
        return f'{self.get_solar_month()}{self.get_name()}'

    def next(self, n: int) -> SolarWeek:
        d: int = self._index
        m: SolarMonth = self.get_solar_month()
        if n > 0:
            d += n
            week_count: int = m.get_week_count(self._start)
            while d >= week_count:
                d -= week_count
                m = m.next(1)
                if m.get_first_day().get_week().get_index() != self._start:
                    d += 1
                week_count = m.get_week_count(self._start)
        elif n < 0:
            d += n
            while d < 0:
                if m.get_first_day().get_week().get_index() != self._start:
                    d -= 1
                m = m.next(-1)
                d += m.get_week_count(self._start)
        return SolarWeek(m.get_year(), m.get_month(), d, self._start)

    def get_first_day(self) -> SolarDay:
        """
        本周第1天
        :return: 公历日
        """
        first_day: SolarDay = SolarDay(self._year, self._month, 1)
        return first_day.next(self._index * 7 - self.index_of(first_day.get_week().get_index() - self._start, 7))

    def get_days(self) -> List[SolarDay]:
        """
        本周公历日列表
        :return: 公历日列表
        """
        l: List[SolarDay] = []
        d: SolarDay = self.get_first_day()
        l.append(d)
        for i in range(1, 7):
            l.append(d.next(i))
        return l

    def __eq__(self, o: SolarWeek) -> bool:
        return o and o.get_first_day() == self.get_first_day()


class SolarDay(DayUnit):
    """
    公历日
    """
    NAMES: List[str] = ['1日', '2日', '3日', '4日', '5日', '6日', '7日', '8日', '9日', '10日', '11日', '12日', '13日', '14日', '15日', '16日', '17日', '18日', '19日', '20日', '21日', '22日', '23日', '24日', '25日', '26日', '27日', '28日', '29日', '30日', '31日']

    @staticmethod
    def validate(year: int, month: int, day: int) -> None:
        if day < 1:
            raise ValueError(f'illegal solar day: {year}-{month}-{day}')
        if 1582 == year and 10 == month:
            if (4 < day < 15) or day > 31:
                raise ValueError(f'illegal solar day: {year}-{month}-{day}')
        elif day > SolarMonth.from_ym(year, month).get_day_count():
            raise ValueError(f'illegal solar day: {year}-{month}-{day}')

    def __init__(self, year: int, month: int, day: int):
        SolarDay.validate(year, month, day)
        super().__init__(year, month, day)

    @classmethod
    def from_ymd(cls, year: int, month: int, day: int) -> SolarDay:
        return cls(year, month, day)

    def get_solar_month(self) -> SolarMonth:
        """
        :return: 公历月
        """
        return SolarMonth.from_ym(self._year, self._month)

    def get_week(self) -> Week:
        """
        :return: 星期
        """
        return self.get_julian_day().get_week()

    def get_constellation(self) -> Constellation:
        """
        :return: 星座 Constellation
        """
        y: int = self._month * 100 + self._day
        return Constellation.from_index(9 if y > 1221 or y < 120 else 10 if y < 219 else 11 if y < 321 else 0 if y < 420 else 1 if y < 521 else 2 if y < 622 else 3 if y < 723 else 4 if y < 823 else 5 if y < 923 else 6 if y < 1024 else 7 if y < 1123 else 8)

    def get_name(self) -> str:
        return self.NAMES[self._day - 1]

    def __str__(self) -> str:
        return f'{self.get_solar_month()}{self.get_name()}'

    def next(self, n: int) -> SolarDay:
        return self.get_julian_day().next(n).get_solar_day()

    def is_before(self, target: SolarDay) -> bool:
        """
        是否在指定公历日之前
        :param target: 公历日
        :return: true/false
        """
        y: int = target.get_year()
        if self._year != y:
            return self._year < y
        m: int = target.get_month()
        return self._month < m if self._month != m else self._day < target.get_day()

    def is_after(self, target: SolarDay) -> bool:
        """
        是否在指定公历日之后
        :param target: 公历日
        :return: true/false
        """
        y: int = target.get_year()
        if self._year != y:
            return self._year > y
        m: int = target.get_month()
        return self._month > m if self._month != m else self._day > target.get_day()

    def get_term(self) -> SolarTerm:
        """
        :return: 当天所在的节气 SolarTerm
        """
        return self.get_term_day().get_solar_term()

    def get_term_day(self) -> SolarTermDay:
        """
        :return: 节气第几天
        """
        y: int = self._year
        i: int = self._month * 2
        if i == 24:
            y += 1
            i = 0
        term: SolarTerm = SolarTerm(y, i)
        day: SolarDay = term.get_solar_day()
        while self.is_before(day):
            term = term.next(-1)
            day = term.get_solar_day()
        return SolarTermDay(term, self.subtract(day))

    def get_solar_week(self, start: int) -> SolarWeek:
        """
        公历周
        :param start: 起始星期，1234560分别代表星期一至星期天
        :return: 公历周
        """
        return SolarWeek(self._year, self._month, ceil((self._day + SolarDay(self._year, self._month, 1).get_week().next(-start).get_index()) / 7.0) - 1, start)

    def get_phenology_day(self) -> PhenologyDay:
        """
        当天所在的七十二候
        :return: 七十二候 PhenologyDay
        """
        d: SolarTermDay = self.get_term_day()
        day_index: int = d.get_day_index()
        index: int = day_index // 5
        if index > 2:
            index = 2
        term: SolarTerm = d.get_solar_term()
        return PhenologyDay(Phenology(term.get_year(), term.get_index() * 3 + index), day_index - index * 5)

    def get_phenology(self) -> Phenology:
        """
        当天所在的候
        :return: 候 Phenology
        """
        return self.get_phenology_day().get_phenology()

    def get_dog_day(self) -> Union[DogDay, None]:
        """
        当天所在的三伏天
        :return: 三伏天 DogDay
        """
        # 夏至
        xia_zhi: SolarTerm = SolarTerm(self._year, 12)
        # 第1个庚日
        start: SolarDay = xia_zhi.get_solar_day()
        # 第3个庚日，即初伏第1天
        start = start.next(start.get_lunar_day().get_sixty_cycle().get_heaven_stem().steps_to(6) + 20)
        days: int = self.subtract(start)
        # 初伏以前
        if days < 0:
            return None
        if days < 10:
            return DogDay(Dog(0), days)

        # 第4个庚日，中伏第1天
        start = start.next(10)
        days = self.subtract(start)
        if days < 10:
            return DogDay(Dog(1), days)

        # 第5个庚日，中伏第11天或末伏第1天
        start = start.next(10)
        days = self.subtract(start)
        # 立秋
        if xia_zhi.next(3).get_solar_day().is_after(start):
            if days < 10:
                return DogDay(Dog(1), days + 10)
            start = start.next(10)
            days = self.subtract(start)
        if days < 10:
            return DogDay(Dog(2), days)
        return None

    def get_plum_rain_day(self) -> Union[PlumRainDay, None]:
        """
        当天所在的梅雨天（芒种后的第1个丙日入梅，小暑后的第1个未日出梅）
        :return: 梅雨天
        """
        # 芒种
        grain_in_ear: SolarTerm = SolarTerm(self._year, 11)
        start: SolarDay = grain_in_ear.get_solar_day()
        # 芒种后的第1个丙日
        start = start.next(start.get_lunar_day().get_sixty_cycle().get_heaven_stem().steps_to(2))
        # 小暑
        end: SolarDay = grain_in_ear.next(2).get_solar_day()
        # 小暑后的第1个未日
        end = end.next(end.get_lunar_day().get_sixty_cycle().get_earth_branch().steps_to(7))
        if self.is_before(start) or self.is_after(end):
            return None
        return PlumRainDay(PlumRain(1), 0) if self == end else PlumRainDay(PlumRain(0), self.subtract(start))

    def get_hide_heaven_stem_day(self) -> HideHeavenStemDay:
        """
        :return: 人元司令分野
        """
        from tyme4py.sixtycycle import HideHeavenStemDay, HideHeavenStem
        day_counts: List[int] = [3, 5, 7, 9, 10, 30]
        term: SolarTerm = self.get_term()
        if term.is_qi():
            term = term.next(-1)
        day_index: int = self.subtract(term.get_solar_day())
        start_index: int = (term.get_index() - 1) * 3
        data: str = '93705542220504xx1513904541632524533533105544806564xx7573304542018584xx95'[start_index: start_index + 6]
        days: int = 0
        heaven_stem_index: int = 0
        type_index: int = 0
        while type_index < 3:
            i: int = type_index * 2
            d: str = data[i: i + 1]
            count: int = 0
            if d != 'x':
                heaven_stem_index = int(d, 10)
                count = day_counts[int(data[i + 1: i + 2], 10)]
                days += count

            if day_index <= days:
                day_index -= days - count
                break

            type_index += 1

        hide_type: HideHeavenStemType
        if type_index == 0:
            hide_type = HideHeavenStemType.RESIDUAL
        elif type_index == 1:
            hide_type = HideHeavenStemType.MIDDLE
        else:
            hide_type = HideHeavenStemType.MAIN
        return HideHeavenStemDay(HideHeavenStem(heaven_stem_index, hide_type), day_index)

    def get_nine_day(self) -> Union[NineDay, None]:
        """
        当天所在的数九天
        :return: 数九天 NineDay
        """
        start: SolarDay = SolarTerm(self._year + 1, 0).get_solar_day()
        if self.is_before(start):
            start = SolarTerm(self._year, 0).get_solar_day()

        end: SolarDay = start.next(81)
        if self.is_before(start) or (not self.is_before(end)):
            return None

        days: int = self.subtract(start)
        return NineDay(Nine(int(days / 9)), days % 9)

    def get_index_in_year(self) -> int:
        """
        位于当年的索引
        :return: 索引
        """
        return self.subtract(SolarDay(self._year, 1, 1))

    def subtract(self, target: SolarDay) -> int:
        """
        公历日期相减，获得相差天数
        :param target: 公历
        :return: 两个公历日之间相差的天数
        """
        return int(self.get_julian_day().subtract(target.get_julian_day()))

    def get_julian_day(self) -> JulianDay:
        """
        公历日转儒略日
        :return: 儒略日
        """
        return JulianDay.from_ymd_hms(self._year, self._month, self._day, 0, 0, 0)

    def get_lunar_day(self) -> LunarDay:
        """
        :return: 农历日
        """
        from tyme4py.lunar import LunarMonth, LunarDay
        m: LunarMonth = LunarMonth.from_ym(self._year, self._month)
        days: int = self.subtract(m.get_first_julian_day().get_solar_day())
        while days < 0:
            m = m.next(-1)
            days += m.get_day_count()
        return LunarDay(m.get_year(), m.get_month_with_leap(), days + 1)

    def get_sixty_cycle_day(self) -> SixtyCycleDay:
        from tyme4py.sixtycycle import SixtyCycleDay
        return SixtyCycleDay.from_solar_day(self)

    def get_rab_byung_day(self) -> RabByungDay:
        from tyme4py.rabbyung import RabByungDay
        return RabByungDay.from_solar_day(self)

    def get_legal_holiday(self) -> Union[LegalHoliday, None]:
        """
        法定假日，如果当天不是法定假日，返回None
        :return: 法定假日
        """
        return LegalHoliday.from_ymd(self._year, self._month, self._day)

    def get_festival(self) -> Union[SolarFestival, None]:
        """
        公历现代节日，如果当天不是公历现代节日，返回None
        :return: 公历现代节日
        """
        return SolarFestival.from_ymd(self._year, self._month, self._day)

    def get_phase_day(self) -> PhaseDay:
        """
        月相第几天
        :return: 月相第几天
        """
        month = self.get_lunar_day().get_lunar_month().next(1)
        p = Phase.from_index(month.get_year(), month.get_month_with_leap(), 0)
        d = p.get_solar_day()
        while d.is_after(self):
            p = p.next(-1)
            d = p.get_solar_day()
        return PhaseDay(p, self.subtract(d))

    def get_phase(self) -> Phase:
        """
        月相
        :return: 月相 Phase。
        """
        return self.get_phase_day().get_phase()


class SolarTime(SecondUnit):

    @staticmethod
    def validate(year: int, month: int, day: int, hour: int, minute: int, second: int) -> None:
        SecondUnit.validate(year, month, day, hour, minute, second)
        SolarDay.validate(year, month, day)

    def __init__(self, year: int, month: int, day: int, hour: int, minute: int, second: int):
        SolarTime.validate(year, month, day, hour, minute, second)
        super().__init__(year, month, day, hour, minute, second)

    @classmethod
    def from_ymd_hms(cls, year: int, month: int, day: int, hour: int, minute: int, second: int) -> SolarTime:
        """
        :param year: 年
        :param month: 月
        :param day: 日
        :param hour: 时
        :param minute: 分
        :param second: 秒
        :return: 公历时刻
        """
        return cls(year, month, day, hour, minute, second)

    def get_solar_day(self) -> SolarDay:
        """
        :return: 公历日
        """
        return SolarDay.from_ymd(self._year, self._month, self._day)

    def get_name(self) -> str:
        return f'{self._hour:02d}:{self._minute:02d}:{self._second:02d}'

    def __str__(self) -> str:
        return f'{self.get_solar_day()} {self.get_name()}'

    def next(self, n: int) -> SolarTime:
        """
        推移
        :param n: 推移秒数
        :return:  公历时刻
        """
        if n == 0:
            return SolarTime(self._year, self._month, self._day, self._hour, self._minute, self._second)
        ts: int = self._second + n
        tm: int = self._minute + floor(ts / 60)
        ts %= 60
        if ts < 0:
            ts += 60
            tm -= 1
        th: int = self._hour + int(tm / 60)
        tm %= 60
        if tm < 0:
            tm += 60
            th -= 1
        td: int = int(th / 24)
        th %= 24
        if th < 0:
            th += 24
            td -= 1
        d: SolarDay = self.get_solar_day().next(td)
        return SolarTime(d.get_year(), d.get_month(), d.get_day(), th, tm, ts)

    def is_before(self, target: SolarTime) -> bool:
        """
        是否在指定公历时刻之前
        :param target: 公历时刻
        :return: true/false
        """
        a_day: SolarDay = self.get_solar_day()
        b_day: SolarDay = target.get_solar_day()
        if a_day != b_day:
            return a_day.is_before(b_day)
        h: int = target.get_hour()
        if self._hour != h:
            return self._hour < h
        m: int = target.get_minute()
        return self._minute < m if self._minute != m else self._second < target.get_second()

    def is_after(self, target: SolarTime) -> bool:
        """
        是否在指定公历时刻之后
        :param target: 公历时刻
        :return: true/false
        """
        a_day: SolarDay = self.get_solar_day()
        b_day: SolarDay = target.get_solar_day()
        if a_day != b_day:
            return a_day.is_after(b_day)
        h: int = target.get_hour()
        if self._hour != h:
            return self._hour > h
        m: int = target.get_minute()
        return self._minute > m if self._minute != m else self._second > target.get_second()

    def get_term(self) -> SolarTerm:
        """
        当时所在的节气
        :return: 节气 SolarTerm
        """
        term: SolarTerm = self.get_solar_day().get_term()
        if self.is_before(term.get_julian_day().get_solar_time()):
            term = term.next(-1)
        return term

    def get_phenology(self) -> Phenology:
        """
        当时所在的候
        :return: 候 Phenology
        """
        p: Phenology = self.get_solar_day().get_phenology()
        if self.is_before(p.get_julian_day().get_solar_time()):
            p = p.next(-1)
        return p

    def get_julian_day(self) -> JulianDay:
        """
        公历时刻转儒略日
        :return: 儒略日
        """
        return JulianDay.from_ymd_hms(self._year, self._month, self._day, self._hour, self._minute, self._second)

    def subtract(self, target: SolarTime) -> int:
        """
        公历时刻相减，获得相差秒数
        :param target: 公历时刻
        :return: 秒数
        """
        days: int = self.get_solar_day().subtract(target.get_solar_day())
        cs: int = self._hour * 3600 + self._minute * 60 + self._second
        ts: int = target.get_hour() * 3600 + target.get_minute() * 60 + target.get_second()
        seconds: int = cs - ts
        if seconds < 0:
            seconds += 86400
            days -= 1
        seconds += days * 86400
        return seconds

    def get_lunar_hour(self) -> LunarHour:
        """
        农历时辰
        :return: 农历时辰
        """
        from tyme4py.lunar import LunarDay, LunarHour
        d: LunarDay = self.get_solar_day().get_lunar_day()
        return LunarHour(d.get_year(), d.get_month(), d.get_day(), self._hour, self._minute, self._second)

    def get_sixty_cycle_hour(self) -> SixtyCycleHour:
        """
        干支时辰
        :return: 干支时辰
        """
        from tyme4py.sixtycycle import SixtyCycleHour
        return SixtyCycleHour.from_solar_time(self)

    def get_phase(self) -> Phase:
        """
        月相
        :return: 月相 Phase。
        """
        month = self.get_lunar_hour().get_lunar_day().get_lunar_month().next(1)
        p = Phase.from_index(month.get_year(), month.get_month_with_leap(), 0)
        while p.get_solar_time().is_after(self):
            p = p.next(-1)
        return p



# -*- coding:utf-8 -*-
from __future__ import annotations

from typing import List, Union

from tyme4py import LoopTyme


class TenStar(LoopTyme):
    """十神"""
    NAMES: List[str] = ['比肩', '劫财', '食神', '伤官', '偏财', '正财', '七杀', '正官', '偏印', '正印']
    """名称"""

    def __init__(self, index_or_name: Union[int, str]):
        super().__init__(self.NAMES, index_or_name)

    @classmethod
    def from_name(cls, name: str) -> TenStar:
        return cls(name)

    @classmethod
    def from_index(cls, index: int) -> TenStar:
        return cls(index)

    def next(self, n: int) -> TenStar:
        return TenStar(self.next_index(n))



# -*- coding:utf-8 -*-
from __future__ import annotations

from typing import TYPE_CHECKING, Union, List, Dict

from tyme4py import AbstractTyme
from tyme4py.culture import Element, Zodiac

if TYPE_CHECKING:
    from tyme4py.sixtycycle import SixtyCycle
    from tyme4py.solar import SolarYear, SolarDay


class RabByungElement(Element):
    """藏历五行"""
    def __init__(self, index_or_name: Union[int, str]):
        super().__init__(index_or_name if isinstance(index_or_name, int) else index_or_name.replace('铁', '金'))

    @classmethod
    def from_index(cls, index: int) -> RabByungElement:
        return cls(index)

    @classmethod
    def from_name(cls, name: str) -> RabByungElement:
        return cls(name)

    def next(self, n: int) -> RabByungElement:
        return RabByungElement.from_index(self.next_index(n))

    def get_reinforce(self) -> RabByungElement:
        """我生者"""
        return self.next(1)

    def get_restrain(self) -> RabByungElement:
        """我克者"""
        return self.next(2)

    def get_reinforced(self) -> RabByungElement:
        """生我者"""
        return self.next(-1)

    def get_restrained(self) -> RabByungElement:
        """克我者"""
        return self.next(-2)

    def __str__(self) -> str:
        return super().__str__().replace('金', '铁')


class RabByungYear(AbstractTyme):
    """藏历年(公历1027年为藏历元年，第一饶迥火兔年）"""
    def __init__(self, rab_byung_index: int, sixty_cycle: SixtyCycle):
        if rab_byung_index < 0 or rab_byung_index > 150:
            raise ValueError(f'illegal rab-byung index: {rab_byung_index}')
        self.rab_byung_index = rab_byung_index
        self.sixty_cycle = sixty_cycle

    @classmethod
    def from_sixty_cycle(cls, rab_byung_index: int, sixty_cycle: SixtyCycle) -> RabByungYear:
        return cls(rab_byung_index, sixty_cycle)

    @classmethod
    def from_element_zodiac(cls, rab_byung_index: int, element: RabByungElement, zodiac: Zodiac) -> RabByungYear:
        from tyme4py.sixtycycle import SixtyCycle
        for i in range(60):
            sc: SixtyCycle = SixtyCycle.from_index(i)
            if sc.get_earth_branch().get_zodiac().get_name() == zodiac.get_name() and sc.get_heaven_stem().get_element().get_index() == element.get_index():
                return cls(rab_byung_index, sc)
        raise ValueError(f'illegal rab-byung element {element}, zodiac {zodiac}')

    @classmethod
    def from_year(cls, year: int) -> RabByungYear:
        from tyme4py.sixtycycle import SixtyCycle
        return cls((year - 1024) // 60, SixtyCycle.from_index(year - 4))

    def get_rab_byung_index(self) -> int:
        """饶迥序号（从0开始）"""
        return self.rab_byung_index

    def get_sixty_cycle(self) -> SixtyCycle:
        """干支"""
        return self.sixty_cycle

    def get_zodiac(self) -> Zodiac:
        """生肖"""
        return self.sixty_cycle.get_earth_branch().get_zodiac()

    def get_element(self) -> RabByungElement:
        """五行"""
        return RabByungElement.from_index(self.sixty_cycle.get_heaven_stem().get_element().get_index())

    def get_name(self) -> str:
        """完整名称"""
        digits: List[str] = ['零', '一', '二', '三', '四', '五', '六', '七', '八', '九']
        units: List[str] = ['', '十', '百']
        n: int = self.rab_byung_index + 1
        s: str = ''
        pos: int = 0

        while n > 0:
            digit: int = n % 10
            if digit > 0:
                s = digits[digit] + units[pos] + s
            elif s:
                s = digits[digit] + s
            n //= 10
            pos += 1

        if s.startswith('一十'):
            s = s.replace('一十', '十', 1)
        return f'第{s}饶迥{self.get_element()}{self.get_zodiac()}年'

    def next(self, n: int) -> RabByungYear:
        return RabByungYear.from_year(self.get_year() + n)

    def get_year(self) -> int:
        """公历年份"""
        return 1024 + self.rab_byung_index * 60 + self.sixty_cycle.get_index()

    def get_leap_month(self) -> int:
        """闰月（闰月数字，1代表闰1月，0代表无闰月）"""
        y: int = 1
        m: int = 4
        t: int = 0
        current_year: int = self.get_year()

        while y < current_year:
            i: int = m - 1 + (33 if t % 2 == 0 else 32)
            y = (y * 12 + i) // 12
            m = i % 12 + 1
            t += 1

        return m if y == current_year else 0

    def get_solar_year(self) -> SolarYear:
        """对应的公历年"""
        from tyme4py.solar import SolarYear
        return SolarYear.from_year(self.get_year())

    def get_first_month(self) -> RabByungMonth:
        """首月"""
        return RabByungMonth(self, 1)

    def get_month_count(self) -> int:
        """月份数量"""
        return 12 if self.get_leap_month() < 1 else 13

    def get_months(self) -> List[RabByungMonth]:
        """所有月份列表"""
        months: List[RabByungMonth] = []
        leap_month: int = self.get_leap_month()
        for i in range(1, 13):
            months.append(RabByungMonth(self, i))
            if i == leap_month:
                months.append(RabByungMonth(self, -i))
        return months

    def __str__(self) -> str:
        return self.get_name()


class RabByungMonth(AbstractTyme):
    """藏历月，仅支持藏历1950年十二月至藏历2050年十二月"""
    NAMES = ['正月', '二月', '三月', '四月', '五月', '六月', '七月', '八月', '九月', '十月', '十一月', '十二月']
    ALIAS = ['神变月', '苦行月', '具香月', '萨嘎月', '作净月', '明净月', '具醉月', '具贤月', '天降月', '持众月', '庄严月', '满意月']
    DAYS: Dict[int, List[int]] = {}
    _isInit: bool = False

    # 初始化特殊日数据（1950-2050年）
    @staticmethod
    def _init():
        if RabByungMonth._isInit:
            return

        years: List[str] = "2c>,182[>1:2TA4ZI=n1E2Bk1J2Ff3Mk503Oc62g=,172^>1:2XA1>2UE2Bo1I2Fj3Lo62Fb3Mf5,03N^72b=1:2]A1>2ZF1B2VI2Em1K2Fe,2Lh1R3Na603P\\:172Y>1;2UB2=m2Dq1J2Eh,2Kl1Q3Me603Pa:172^>1;2YA2=p1C2UI,2Dk2Jp3QEc3Mi603Pf:3L[72b?1:2]A1<2UB2XH,2Cn1I2Ei1L2Ie1Q3Na703Q\\:2`@1;2XA,4\\H;m1B2TI2Em1L2Ij1Q3Nf603Q`903QW:,2[@1;2TB2XI1E4TMAh2Io3RFe3Mj603Pc803Q[;,2^?1;2WA2>q1E2Bm1I2Fi1M2Hc3Of70,3P^82a>1:2[A1>2WE1B2TI2Fm1L2Hf3Ni6,03Oa703PZ:3`A62V>4]F;q1B4YJ>l2Eq1L2Gi3Ml5,03Nd603Q_9172[>1;2XB2>p1E2VK2Fl,1K2Fc3Mh603Pc9172`>1;2\\B1>2UD2=j2En,1J2Fg3Mm62Ib3Pj;3M_703R[:2`B1=2YB2=n,1C2TI2Fk1L2Ig1P3Nd703Q_:152X<2[A,2<q1B2WI2Ep1L2Il1Q3Ni703Qc9152[:2^@,1;2WB2>o1E2Bk1I2Fh1M2Ib3Pf803R^9,2a?1;2ZA1>2UE2Bp1I2Fl1M2If3Oi80,3Pa803QY:2^A1>2ZE1B4WJ>j2Fp1M2Hi1N2H`,3Od703Q]:162Y>1;2VB2?o1E4VM@h2Gl1M,2Hd3Ng603Qa9172^>1;2ZB1?2UE2@l2Fo1L,2Gg3Mk62H`3Pf:172c?3QY;2_B1>2YD2?o1E,2TK2Fj1M2Ie1P3Mb703R^;172X=2\\C1>,2TD2WJ2Fn1L2Ij1P3Ng703Rb:162[<2_B1=,2VC2>m1E4TMAh2Io3QFe3Nl82Ja3Qf:152_;0,3RU<2ZB1>2TE2Bn1I2Fj1M2Je3Pk:2K^3Ra:,03RY;2]A1>2XE1B2TI2Fo1M2Ii1P2Ka3Qd8,03R]:3bB62W>4]F:q1B2?n1F4VNAh2Il1O2Jd,3Pg803Q`:162\\=1;2XB1?2TF2Bl2Ho1N,2Ig3Nk703Qd9162`>1;2]B1?2XE2Ao1G2TM,2Hj1M2Id1P3M_603R\\;172W>2\\E1@2TE,2?i2Gm1M2Ih1P3Md603Ra;172[=28q1?2WD,2?m2Fq1M2Il1P3Mi72I^3Re:162_<172W=,2ZC2?q1E2Bk1I2Fh1M2Jd1Q3M^52b;16,2Y<2]B1>2VE2Bp1I2Fm1M2Jh1Q2Lb3Re:15,2\\;3aC62U>2[E1B4WJ>k1F4TNBg2Jl1P2Le3Qh9,03R`:172Z=1:2VB2?q1F2Bk2Ip1P2Jg,1P2J_3Qc:162^=1;2[B1?2WF2Bo1H2Bg2Ij,1O2Jc3Qg:3L\\62c>3QY;3aC72V?2[F1A2TG2Bj,2Hm1N2Jg1P3Mb603R_;182Z>1:2T@2WF2Am,2Gp1M2Ik1P3Mg603Rc;172^>192W?2ZE,2@p1F2Bj2Io3QEe1M2Jb1Q3M]72b=182Z>,2]D1?2VE2Bn1I2Fk1M2Jg1Q3Ma62e<172]=,172U>2YE1B2UI2Fp1N2Jk1Q3Me503M\\6,2`<172Y>3_F:2TB2?n1F2Cj2Jo3QDc2Lh1R,3L_52c;172]=1:2XB1?2UF2Cn1I2Eg2Kk1P,2Lb3Rf;162a=1:2]B1?2ZF1B2TH2Dj2Jm,1O2Kf1Q3M`603Q\\;182Y?2;q1A2WH2Cm,2Hq1O2Ji1P3Me603Qa;182]>1:2WA2[G2Ap,1G2Bi2Im1P3Mi72I_3Qf;3N\\72Eh1:2Z?29o,1@2UF2Bm1I2Fh1M2Je1Q3N`72f?3PY92]>19,2U?2YF2Bq1I2Fm1M2Jj1Q3Nd603O]72`=,182X?4]F:o1B4WI=k1F4UNCi2Jn3REc3Mh503N`6,2c<182\\>1:2VA2?q1F2Cm1J2Fg2Lk1R3Mc5,2f<172`=1:2[A1?2XF2Cq1I2Ek2Kn1R,2Lf1R3N_62d>3PZ:3aC72W?2;p1B2WI2Dn1J,2De2Ki1Q3Mc603Q_:182\\?1;2VB2<m2Cq1I,2Dh2Jl1P3Mg603Qd;182`?1;2ZA2<p1B,2UH2Cl1I2Ef3Mm82Jc1Q3N_703QY:2]@1;2UA,2XG2Bp1I2Fk1M2Jh1Q3Nc703Q]92`?1:,2X@4\\G:n1B2VI2Fp1M2Jl1R3Ng603P`82d>,192[?1;2UA2>o1F2Ck1J2Gg3Mk603Oc70,3OZ82_>1:2YA1?2VF2Cp1J2Fj1M2Gc3Nf5,03O^72b>1:2^B1?4[G;n1C2VJ2Fn1L2Gf,3Mi503Nb603Q]:172Y?1<2UB2>m2Eq1K2Fi,2Kl1R3Mf603Qa:182^?1;2YB2>q1D2VJ,2Dl1J2Fe3Mj603Qg;3N]72c@3QX;2]A1=2VB,2YI2Co1J2Fi1M2Je1Q3Nb703R]:2aA1<2XA,2<n1C2UI2Fn1M2Jj1Q3Nf703Q`903RX:,2[@1<2TB4YJ>l1E4UNBi1J2Ge3Mk703Pc803Q[9,2^?1;2XB2>q1E2Cn1J2Gj1M2Ic3Of70,3P^82b?1;2\\A1>2XF1C2UJ2Fm1M2Hf3Ni6,03Oa703Q[:3aB72W>1<2TC2?m2Fq1L2Gi3Ml5,03Ne703Q_:172\\>1<2XB2?q1E2WL2Fl,1L2Gd3Ni603Qd:172a?1;2\\B1>2VD2>k,2Eo1K2Gh1M2Ic1Q3N`703R\\;3aC62U=2YC2>o,1D2TJ2Fl1M2Jh1Q3Ne703R`:162Y<2\\B,1=2TC4XJ=j2Fp1M2Jm3QFc3Ni803Qc:152\\;2_A,1<2WB2>o1E2Bl1J2Gh1N2Jc3Qg903R^:,2b@1;2[B1>2VE2Cq1J2Gl1N2Jf3Pj80,3Qa803RZ;2_B1>4[F:o1C4XK?k2Fp1M2Ii1O2Ia,3Pd703R^:172Y>1<2VC2?p1F2Ai2Hl1M,2Hd3Oh703Qb:172^>1<2[C1?2UE2Al2Go,1L2Hg3Nl82Ia3Qg;3M]72e@3RZ;3`C72T>2YD2@o1E,2TK2Gk1M2Jf1Q3Nb703R^;172Y=2\\D1>,2TD4XK>i2Fo1M2Jj1Q3Ng703Rb;172\\<2`C1=,2WC2?n1F4VNBi1J2Gf1N2Kb3Rf:162_;15,2V<2ZB1?2TE2Bn1J2Gk1N2Kf1Q2L^3Rb:,152Z;2^B1>2YE1B2UJ2Go1N2Ji1P2Kb3Qd9,03R];172X>1;2TC2@n1G2Bi2Im1O2Jd,3Ph803Ra:172\\>1;2YC1@2UF2Bl2Hp1N,2Ig3Ol82J`3Qe:172a>1;4^C7q1?2XF2Ao1G2UN,2Hj1N2Jd1Q3N`703R];182X>2]F1@2TF,2@j2Gn1M2Jq1Q3Ne703Ra;172\\>192T?,2WE2@m1F4TMAf2Im3QEc3Nj82J`3Rf;172_=182W>,2ZD2?q1F2Bl1I2Gj1N2Ke1R3M_62b<17,2Z=2]C1?2WE2Bq1I2Gn1N2Ki1Q3Mb52e;16,2]<172V>4[F:o1B4XK?l1G4UOCh2Jl1Q2Le3Rh:,152`;172Z>1;2WB2@q1G2Cl2Ip1P2K_".split(',')

        y: int = 1950
        m: int = 11
        for ys in years:
            data: str = ys.strip()
            while data:
                length: int = ord(data[0]) - 48
                chunk: str = data[1:1+length]
                data: str = data[1+length:]
                days_data: List[int] = [ord(c) - 53 - 30 for c in chunk]
                RabByungMonth.DAYS[y * 13 + m] = days_data
                m += 1
            y += 1
            m = 0

    def __init__(self, year: RabByungYear, month: int):
        RabByungMonth._init()
        if month == 0 or month > 12 or month < -12:
            raise ValueError(f'illegal rab-byung month: {month}')

        y: int = year.get_year()
        if y < 1950 or y > 2050:
            raise ValueError(f'rab-byung year {y} must between 1950 and 2050')

        m: int = abs(month)
        if y == 1950 and m < 12:
            raise ValueError(f'month {month} must be 12 in rab-byung year {y}')

        leap: bool = month < 0
        leap_month: int = year.get_leap_month()
        if leap and m != leap_month:
            raise ValueError(f'illegal leap month {m}  in rab-byung year {y}')

        self.year = year
        self.month = m
        self.leap = leap

        index: int = m - 1
        if leap or (0 < leap_month < m):
            index += 1
        self.index_in_year = index

    @classmethod
    def from_ym(cls, year: int, month: int) -> RabByungMonth:
        return cls(RabByungYear.from_year(year), month)

    @classmethod
    def from_element_zodiac(cls, rab_byung_index: int, element: RabByungElement, zodiac: Zodiac, month: int) -> RabByungMonth:
        return cls(RabByungYear.from_element_zodiac(rab_byung_index, element, zodiac), month)

    def get_rab_byung_year(self) -> RabByungYear:
        return self.year

    def get_year(self) -> int:
        return self.year.get_year()

    def get_month(self) -> int:
        return self.month

    def get_month_with_leap(self) -> int:
        return -self.month if self.leap else self.month

    def get_index_in_year(self) -> int:
        return self.index_in_year

    def is_leap(self) -> bool:
        return self.leap

    def get_name(self) -> str:
        return ('闰' if self.leap else '') + self.NAMES[self.month - 1]

    def get_alias(self) -> str:
        return ('闰' if self.leap else '') + self.ALIAS[self.month - 1]

    def next(self, n: int) -> RabByungMonth:
        if n == 0:
            return RabByungMonth.from_ym(self.get_year(), self.get_month_with_leap())
        m: int = self.index_in_year + 1 + n
        y: RabByungYear = self.year
        if n > 0:
            month_count: int = y.get_month_count()
            while m > month_count:
                m -= month_count
                y = y.next(1)
                month_count = y.get_month_count()
        else:
            while m <= 0:
                y = y.next(-1)
                m += y.get_month_count()

        leap: bool = False
        leap_month: int = y.get_leap_month()
        if leap_month > 0:
            if m == leap_month + 1:
                leap = True
            if m > leap_month:
                m -= 1
        return RabByungMonth(y, -m if leap else m)

    def get_first_day(self) -> RabByungDay:
        return RabByungDay(self, 1)

    def get_days(self) -> List[RabByungDay]:
        days: List[RabByungDay] = []
        miss_days: List[int] = self.get_miss_days()
        leap_days: List[int] = self.get_leap_days()
        for day in range(1, 31):
            if day in miss_days:
                continue
            days.append(RabByungDay(self, day))
            if day in leap_days:
                days.append(RabByungDay(self, -day))
        return days

    def get_day_count(self) -> int:
        return 30 + len(self.get_leap_days()) - len(self.get_miss_days())

    def get_special_days(self) -> List[int]:
        return self.DAYS.get(self.get_year() * 13 + self.index_in_year, [])

    def get_leap_days(self) -> List[int]:
        return [d for d in self.get_special_days() if d > 0]

    def get_miss_days(self) -> List[int]:
        return [-d for d in self.get_special_days() if d < 0]

    def __str__(self) -> str:
        return f'{self.year}{self.get_name()}'


class RabByungDay(AbstractTyme):
    """藏历日，仅支持藏历1950年十二月初一（公历1951年1月8日）至藏历2050年十二月三十（公历2051年2月11日）"""
    NAMES: List[str] = ['初一', '初二', '初三', '初四', '初五', '初六', '初七', '初八', '初九', '初十', '十一', '十二', '十三', '十四', '十五', '十六', '十七', '十八', '十九', '二十', '廿一', '廿二', '廿三', '廿四', '廿五', '廿六', '廿七', '廿八', '廿九', '三十']

    def __init__(self, month: RabByungMonth, day: int):
        if day == 0 or day < -30 or day > 30:
            raise ValueError(f'illegal day {day} in {month}')

        d: int = abs(day)
        leap: bool = day < 0

        # 验证日期的合法性
        if leap and d not in month.get_leap_days():
            raise ValueError(f'illegal leap day {d} in {month}')
        elif not leap and d in month.get_miss_days():
            raise ValueError(f'illegal day {d} in {month}')

        self.month = month
        self.day = d
        self.leap = leap

    @classmethod
    def from_ymd(cls, year: int, month: int, day: int) -> RabByungDay:
        return cls(RabByungMonth.from_ym(year, month), day)

    @classmethod
    def from_element_zodiac(cls, rab_byung_index: int, element: RabByungElement, zodiac: Zodiac, month: int, day: int) -> RabByungDay:
        return cls(RabByungMonth.from_element_zodiac(rab_byung_index, element, zodiac, month), day)

    @classmethod
    def from_solar_day(cls, solar_day: SolarDay) -> RabByungDay:
        from tyme4py.solar import SolarDay
        days: int = solar_day.subtract(SolarDay.from_ymd(1951, 1, 8))
        m: RabByungMonth = RabByungMonth.from_ym(1950, 12)
        count: int = m.get_day_count()
        while days >= count:
            days -= count
            m = m.next(1)
            count = m.get_day_count()
        day: int = days + 1
        for d in m.get_special_days():
            if d < 0:
                if day >= -d:
                    day += 1
            elif d > 0:
                if day == d + 1:
                    day = -d
                    break
                elif day > d + 1:
                    day -= 1
        return RabByungDay(m, day)

    def get_rab_byung_month(self) -> RabByungMonth:
        return self.month

    def get_year(self) -> int:
        return self.month.get_year()

    def get_month(self) -> int:
        return self.month.get_month_with_leap()

    def get_day(self) -> int:
        return self.day

    def is_leap(self) -> bool:
        return self.leap

    def get_day_with_leap(self) -> int:
        return -self.day if self.leap else self.day

    def get_name(self) -> str:
        return ('闰' if self.leap else '') + self.NAMES[self.day - 1]

    def subtract(self, target: RabByungDay) -> int:
        return self.get_solar_day().subtract(target.get_solar_day())

    def get_solar_day(self) -> SolarDay:
        from tyme4py.solar import SolarDay
        month: RabByungMonth = RabByungMonth.from_ym(1950, 12)
        total_days: int = 0

        # 计算从1950年12月到当前月的总天数
        while not month == self.month:
            total_days += month.get_day_count()
            month = month.next(1)

        day_num: int = self.day
        for d in month.get_special_days():
            if d < 0:
                if day_num > -d:
                    day_num -= 1
            elif d > 0:
                if day_num > d:
                    day_num += 1

        if self.leap:
            day_num += 1

        return SolarDay.from_ymd(1951, 1, 7).next(total_days + day_num)

    def next(self, n: int) -> RabByungDay:
        return self.get_solar_day().next(n).get_rab_byung_day()

    def __str__(self) -> str:
        return f'{self.month}{self.get_name()}'



# -*- coding:utf-8 -*-
from __future__ import annotations

from abc import ABC

from tyme4py import AbstractTyme


class YearUnit(AbstractTyme, ABC):
    """年"""

    _year: int

    def __init__(self, year: int):
        self._year = year

    def get_year(self) -> int:
        return self._year


class MonthUnit(YearUnit, ABC):
    """月"""

    _month: int

    def __init__(self, year: int, month: int):
        super().__init__(year)
        self._month = month

    def get_month(self) -> int:
        return self._month


class DayUnit(MonthUnit, ABC):
    """日"""

    _day: int

    def __init__(self, year: int, month: int, day: int):
        super().__init__(year, month)
        self._day = day

    def get_day(self) -> int:
        return self._day


class SecondUnit(DayUnit, ABC):
    """秒"""

    _hour: int
    _minute: int
    _second: int

    def __init__(self, year: int, month: int, day: int, hour: int, minute: int, second: int):
        super().__init__(year, month, day)
        self._hour = hour
        self._minute = minute
        self._second = second

    def get_hour(self) -> int:
        return self._hour

    def get_minute(self) -> int:
        return self._minute

    def get_second(self) -> int:
        return self._second

    @staticmethod
    def validate(year: int, month: int, day: int, hour: int, minute: int, second: int) -> None:
        if hour < 0 or hour > 23:
            raise ValueError(f'illegal hour: {hour}')
        if minute < 0 or minute > 59:
            raise ValueError(f'illegal minute: {minute}')
        if second < 0 or second > 59:
            raise ValueError(f'illegal second: {second}')


class WeekUnit(MonthUnit, ABC):
    """周"""

    _index: int
    _start: int

    def __init__(self, year: int, month: int, index: int, start: int):
        super().__init__(year, month)
        self._index = index
        self._start = start

    def get_index(self) -> int:
        return self._index

    def get_start(self) -> int:
        return self._start

    @staticmethod
    def validate(year: int, month: int, index: int, start: int) -> None:
        if index < 0 or index > 5:
            raise ValueError(f'illegal week index: {index}')
        if start < 0 or start > 6:
            raise ValueError(f'illegal week start: {start}')



# -*- coding:utf-8 -*-
import unittest

from tyme4py.solar import SolarDay


class TestPlumRainDay(unittest.TestCase):
    def test0(self):
        d = SolarDay(2024, 6, 10).get_plum_rain_day()
        assert d is None

    def test1(self):
        d = SolarDay(2024, 6, 11).get_plum_rain_day()
        assert d
        assert d.get_name() == '入梅'
        assert d.get_plum_rain().__str__() == '入梅'
        assert d.__str__() == '入梅第1天'

    def test2(self):
        d = SolarDay(2024, 7, 6).get_plum_rain_day()
        assert d
        assert d.get_name() == '出梅'
        assert d.get_plum_rain().__str__() == '出梅'
        assert d.__str__() == '出梅'

    def test3(self):
        d = SolarDay(2024, 7, 5).get_plum_rain_day()
        assert d
        assert d.get_name() == '入梅'
        assert d.get_plum_rain().__str__() == '入梅'
        assert d.__str__() == '入梅第25天'



# -*- coding:utf-8 -*-
import unittest

from tyme4py.solar import SolarTerm, SolarDay


class TestSolarTerm(unittest.TestCase):
    def test0(self):
        # 冬至在去年，2022 - 12 - 22 05: 48:11
        dong_zhi = SolarTerm(2023, '冬至')
        assert dong_zhi.get_name() == '冬至'
        assert dong_zhi.get_index() == 0
        # 公历日
        assert dong_zhi.get_julian_day().get_solar_day().__str__() == '2022年12月22日'
        assert dong_zhi.get_solar_day().__str__() == '2022年12月22日'

        # 冬至顺推23次，就是大雪 2023 - 12 - 07 17: 32:55
        daXue = dong_zhi.next(23)
        assert daXue.get_name() == '大雪'
        assert daXue.get_index() == 23
        assert daXue.get_julian_day().get_solar_day().__str__() == '2023年12月7日'
        assert daXue.get_solar_day().__str__() == '2023年12月7日'

        # 冬至逆推2次，就是上一年的小雪 2022 - 11 - 22 16: 20:28
        xiaoXue = dong_zhi.next(-2)
        assert xiaoXue.get_name() == '小雪'
        assert xiaoXue.get_index() == 22
        assert xiaoXue.get_julian_day().get_solar_day().__str__() == '2022年11月22日'
        assert xiaoXue.get_solar_day().__str__() == '2022年11月22日'

        # 冬至顺推24次，就是下一个冬至 2023 - 12 - 22 11: 27:20
        dong_zhi2 = dong_zhi.next(24)
        assert dong_zhi2.get_name() == '冬至'
        assert dong_zhi2.get_index() == 0
        assert dong_zhi2.get_julian_day().get_solar_day().__str__() == '2023年12月22日'
        assert dong_zhi2.get_solar_day().__str__() == '2023年12月22日'

    def test1(self):
        # 公历2023年的雨水，2023 - 02 - 19 06: 34:16
        jq = SolarTerm(2023, '雨水')
        assert jq.get_name() == '雨水'
        assert jq.get_index() == 4

    def test2(self):
        # 公历2023年的大雪，2023 - 12 - 07 17: 32:55
        jq = SolarTerm(2023, '大雪')
        assert jq.get_name() == '大雪'
        # 索引
        assert jq.get_index() == 23
        # 公历
        assert jq.get_julian_day().get_solar_day().__str__() == '2023年12月7日'
        assert jq.get_solar_day().__str__() == '2023年12月7日'
        # 农历
        assert jq.get_julian_day().get_solar_day().get_lunar_day().__str__() == '农历癸卯年十月廿五'
        # 推移
        assert jq.next(5).get_name() == '雨水'

    def test3(self):
        assert SolarDay(2023, 10, 10).get_term().get_name() == '寒露'

    def test4(self):
        assert SolarDay(2023, 12, 7).get_term_day().__str__() == '大雪第1天'
        assert SolarDay(2023, 12, 7).get_term_day().get_day_index() == 0

        assert SolarDay(2023, 12, 8).get_term_day().__str__() == '大雪第2天'
        assert SolarDay(2023, 12, 21).get_term_day().__str__() == '大雪第15天'

        assert SolarDay(2023, 12, 22).get_term_day().__str__() == '冬至第1天'

    def test5(self):
        assert SolarTerm.from_name(1034, '寒露').get_solar_day().__str__() == '1034年10月1日'
        assert SolarTerm.from_name(1034, '寒露').get_julian_day().get_solar_day().__str__() == '1034年10月3日'
        assert SolarTerm.from_name(1034, '寒露').get_julian_day().get_solar_time().__str__() == '1034年10月3日 06:02:28'



# -*- coding:utf-8 -*-
from __future__ import annotations

from typing import List, Union

from tyme4py import LoopTyme
from tyme4py.culture import Direction, Element


class Dipper(LoopTyme):
    """北斗九星"""
    NAMES: List[str] = ['天枢', '天璇', '天玑', '天权', '玉衡', '开阳', '摇光', '洞明', '隐元']
    """名称"""

    def __init__(self, index_or_name: Union[int, str]):
        super().__init__(self.NAMES, index_or_name)

    @classmethod
    def from_name(cls, name: str) -> Dipper:
        return cls(name)

    @classmethod
    def from_index(cls, index: int) -> Dipper:
        return cls(index)

    def next(self, n: int) -> Dipper:
        return Dipper(self.next_index(n))


class NineStar(LoopTyme):
    """九星"""
    NAMES: List[str] = ['一', '二', '三', '四', '五', '六', '七', '八', '九']
    """名称"""

    def __init__(self, index_or_name: Union[int, str]):
        super().__init__(self.NAMES, index_or_name)

    @classmethod
    def from_name(cls, name: str) -> NineStar:
        return cls(name)

    @classmethod
    def from_index(cls, index: int) -> NineStar:
        return cls(index)

    def next(self, n: int) -> NineStar:
        return NineStar(self.next_index(n))

    def get_color(self) -> str:
        """
        颜色
        :return: 颜色
        """
        return ['白', '黑', '碧', '绿', '黄', '白', '赤', '白', '紫'][self.get_index()]

    def get_element(self) -> Element:
        """
        五行
        :return: 五行
        """
        return Element([4, 2, 0, 0, 2, 3, 3, 2, 1][self.get_index()])

    def get_dipper(self) -> Dipper:
        """
        北斗九星
        :return: 北斗九星
        """
        return Dipper(self.get_index())

    def get_direction(self) -> Direction:
        """
        方位
        :return: 方位
        """
        return Direction(self.get_index())

    def __str__(self) -> str:
        return f'{self.get_name()}{self.get_color()}{self.get_element().get_name()}'



# -*- coding:utf-8 -*-
from __future__ import annotations

import warnings
from math import floor, ceil
from typing import TYPE_CHECKING, List, Union

from tyme4py import LoopTyme
from tyme4py.culture import Twenty, Direction, KitchenGodSteed, Week, Duty, Phase, God, Taboo, PhaseDay
from tyme4py.culture.star.twentyeight import TwentyEightStar
from tyme4py.culture.fetus import FetusMonth, FetusDay
from tyme4py.culture.ren import MinorRen
from tyme4py.culture.star.nine import NineStar
from tyme4py.culture.star.six import SixStar
from tyme4py.culture.star.twelve import TwelveStar
from tyme4py.eightchar import EightChar
from tyme4py.eightchar.provider import EightCharProvider
from tyme4py.eightchar.provider.impl import DefaultEightCharProvider
from tyme4py.festival import LunarFestival
from tyme4py.jd import JulianDay
from tyme4py.unit import YearUnit, MonthUnit, WeekUnit, DayUnit, SecondUnit
from tyme4py.util import ShouXingUtil

if TYPE_CHECKING:
    from tyme4py.solar import SolarDay, SolarTime, SolarTerm
    from tyme4py.sixtycycle import SixtyCycle, SixtyCycleDay, SixtyCycleHour, ThreePillars


class LunarYear(YearUnit):
    """
    农历年
    依据国家标准《农历的编算和颁行》GB/T 33661-2017，农历年以正月初一开始，至除夕结束。
    """
    _isInit: bool = False
    _LEAP: List[List[int]] = []

    @staticmethod
    def _init():
        if LunarYear._isInit:
            return
        chars: str = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_@'
        months: List[str] = [
            '080b0r0j0j0j0C0j0j0C0j0j0j0C0j0C0j0C0F0j0V0V0V0u0j0j0C0j0j0j0j0V0C0j1v0u0C0V1v0C0b080u110u0C0j0C1v9K1v2z0j1vmZbl1veN3s1v0V0C2S1v0V0C2S2o0C0j1Z1c2S1v0j1c0j2z1v0j1c0j392H0b2_2S0C0V0j1c0j2z0C0C0j0j1c0j0N250j0C0j0b081n080b0C0C0C1c0j0N',
            '0r1v1c1v0V0V0F0V0j0C0j0C0j0V0j0u1O0j0C0V0j0j0j0V0b080u0r0u080b0j0j0C0V0C0V0j0b080V0u080b0j0j0u0j1v0u080b1c0j080b0j0V0j0j0V0C0N1v0j1c0j0j1v2g1v420j1c0j2z1v0j1v5Q9z1v4l0j1vfn1v420j9z4l1v1v2S1c0j1v2S3s1v0V0C2S1v1v2S1c0j1v2S2_0b0j2_2z0j1c0j',
            '0z0j0j0j0C0j0j0C0j0j0j0C0j0C0j0j0j0j0m0j0C0j0j0C0j0j0j0j0b0V0j0j0C0j0j0j0j0V0j0j0j0V0b0V0V0C0V0C0j0j0b080u110u0V0C0j0N0j0b080b080b0j0r0b0r0b0j0j0j0j0C0j0b0r0C0j0b0j0C0C0j0j0j0j0j0j0j0j0j0b110j0b0j0j0j0C0j0C0j0j0j0j0b080b080b0V080b080b0j0j0j0j0j0j0V0j0j0u1v0j0j0j0C0j0j0j0V0C0N1c0j0C0C0j0j0j1n080b0j0V0C0j0C0C2g0j1c0j0j1v2g1v0j0j1v7N0j1c0j3L0j0j1v5Q1Z5Q1v4lfn1v420j1v5Q1Z5Q1v4l1v2z1v',
            '0H140r0N0r140r0u0r0V171c11140C0j0u110j0u0j1v0j0C0j0j0j0b080V0u080b0C1v0j0j0j0C0j0b080V0j0j0b080b0j0j0j0j0b080b0C080j0b080b0j0j0j0j0j0j0b080j0b080C0b080b080b080b0j0j0j0j080b0j0C0j0j0j0b0j0j080C0b0j0j0j0j0j0j0b08080b0j0C0j0j0j0b0j0j0K0b0j0C0j0j0j0b080b080j0C0b0j080b080b0j0j0j0j080b0j0b0r0j0j0j0b0j0C0r0b0j0j0j0j0j0j0j0b080j0b0r0C0j0b0j0j0j0r0b0j0C0j0j0j0u0r0b0C0j080b0j0j0j0j0j0j0j1c0j0b0j0j0j0C0j0j0j0j0j0j0j0b080j1c0u0j0j0j0C0j1c0j0u0j1c0j0j0j0j0j0j0j0j1c0j0u1v0j0j0V0j0j2g0j0j0j0C1v0C1G0j0j0V0C1Z1O0j0V0j0j2g1v0j0j0V0C2g5x1v4l1v421O7N0V0C4l1v2S1c0j1v2S2_',
            '050b080C0j0j0j0C0j0j0C0j0j0j0C0j0C0j0C030j0j0j0j0j0j0j0j0j0C0j0b080u0V080b0j0j0V0j0j0j0j0j0j0j0j0j0V0N0j0C0C0j0j0j0j0j0j0j0j1c0j0u0j1v0j0j0j0j0j0b080b080j0j0j0b080b080b080b080b0j0j0j080b0j0b080j0j0j0j0b080b0j0j0r0b080b0b080j0j0j0j0b080b080j0b080j0b080b080b080b080b0j0j0r0b0j0b080j0j0j0j0b080b0j0j0C080b0b080j0j0j0j0j0j0j0b080u080j0j0b0j0j0j0C0j0b080j0j0j0j0b080b080b080b0C080b080b080b0j0j0j0j0j0j0b0C080j0j0b0j0j0j0C0j0b080j0j0C0b080b080j0b0j0j0C080b0j0j0j0j0j0j0b0j0j080C0b0j080b0j0j0j0j0j0j0j0C0j0j0j0b0j0j0C080b0j0j0j0j0j0j0b080b080b0K0b080b080b0j0j0j0j0j0j0j0C0j0j0u0j0j0V0j080b0j0C0j0j0j0b0j0r0C0b0j0j0j0j0j0j0j0j0j0C0j0b080b080b0j0C0C0j0C0j0j0j0u110u0j0j0j0j0j0j0j0j0C0j0j0u0j1c0j0j0j0j0j0j0j0j0V0C0u0j0C0C0V0C1Z0j0j0j0C0j0j0j1v0u0j1c0j0j0j0C0j0j2g0j1c1v0C1Z0V0j4l0j0V0j0j2g0j1v0j1v2S1c7N1v',
            '0w0j1c0j0V0j0j0V0V0V0j0m0V0j0C1c140j0j0j0C0V0C0j1v0j0N0j0C0j0j0j0V0j0j1v0N0j0j0V0j0j0j0j0j0j080b0j0j0j0j0j0j0j080b0j0C0j0j0j0b0j0j080u080b0j0j0j0j0j0j0b080b080b080C0b0j080b080b0j0j0j0j080b0j0C0j0j0j0b0j0j080u080b0j0j0j0j0j0j0b080b080b080b0r0b0j080b080b0j0j0j0j080b0j0b0r0j0j0b080b0j0j080b0j080b0j080b080b0j0j0j0j0j0b080b0r0C0b080b0j0j0j0j080b0b080b080j0j0j0b080b080b080b0j0j0j0j080b0j0b080j0j0j0j0b080b0j0j0r0b080b0j0j0j0j0j0b080b080j0b0r0b080j0b080b0j0j0j0j080b0j0b080j0j0j0j0b080b0j080b0r0b0j080b080b0j0j0j0j0j0b080b0r0C0b080b0j0j0j0j0j0j0b080j0j0j0b080b080b080b0j0j0j0r0b0j0b080j0j0j0j0b080b0r0b0r0b0j080b080b0j0j0j0j0j0j0b0r0j0j0j0b0j0j0j0j080b0j0b080j0j0j0j0b080b080b0j0r0b0j080b0j0j0j0j0j0j0j0b0r0C0b0j0j0j0j0j0j0j080b0j0C0j0j0j0b0j0C0r0b0j0j0j0j0j0j0b080b080u0r0b0j080b0j0j0j0j0j0j0j0b0r0C0u0j0j0j0C0j080b0j0C0j0j0j0u110b0j0j0j0j0j0j0j0j0j0C0j0b080b0j0j0C0C0j0C0j0j0j0b0j1c0j080b0j0j0j0j0j0j0V0j0j0u0j1c0j0j0j0C0j0j2g0j0j0j0C0j0j0V0j0b080b1c0C0V0j0j2g0j0j0V0j0j1c0j1Z0j0j0C0C0j1v',
            '160j0j0V0j1c0j0C0j0C0j1f0j0V0C0j0j0C0j0j0j1G080b080u0V080b0j0j0V0j1v0j0u0j1c0j0j0j0C0j0j0j0C0C0j1D0b0j080b0j0j0j0j0C0j0b0r0C0j0b0j0C0C0j0j0j0j0j0j0j0j0j0b0r0b0r0j0b0j0j0j0C0j0b0r0j0j0j0b080b080j0b0C0j080b080b0j0j0j0j0j0j0b0C080j0j0b0j0j0j0C0j0b080j0j0j0j0b080b080j0b0C0r0j0b0j0j0j0j0j0j0b0C080j0j0b0j0j0j0C0j0j0j0j0C0j0j0b080b0j0j0C080b0j0j0j0j0j0j0b080b080b080C0b080b080b080b0j0j0j0j0j0b080C0j0j0b080b0j0j0C080b0j0j0j0j0j0j0b080j0b0C080j0j0b0j0j0j0j0j0j0b080j0b080C0b080b080b080b0j0j0j0j080b0j0C0j0j0b080b0j0j0C080b0j0j0j0j0j0j0b080j0b080u080j0j0b0j0j0j0j0j0j0b080C0j0j0b080b0j0j0C0j0j080b0j0j0j0j0j0b080b0C0r0b080b0j0j0j0j0j0j0b080j0b080u080b080b080b0j0j0j0C0j0b080j0j0j0j0b0j0j0j0C0j0j080b0j0j0j0j0j0b080b0C0r0b080b0j0j0j0j0j0j0b080j0b0r0b080b080b080b0j0j0j0r0b0j0b0r0j0j0j0b0j0j0j0r0b0j080b0j0j0j0j0j0j0j0b0r0C0b0j0j0j0j0j0j0j0b080j0C0u080b080b0j0j0j0r0b0j0C0C0j0b0j110b0j080b0j0j0j0j0j0j0u0r0C0b0j0j0j0j0j0j0j0j0j0C0j0j0j0b0j1c0j0C0j0j0j0b0j0814080b080b0j0j0j0j0j0j1c0j0u0j0j0V0j0j0j0j0j0j0j0u110u0j0j0j',
            '020b0r0C0j0j0j0C0j0j0V0j0j0j0j0j0C0j1f0j0C0j0V1G0j0j0j0j0V0C0j0C1v0u0j0j0j0V0j0j0C0j0j0j1v0N0C0V0j0j0j0K0C250b0C0V0j0j0V0j0j2g0C0V0j0j0C0j0j0b081v0N0j0j0V0V0j0j0u0j1c0j080b0j0j0j0j0j0j0V0j0j0u0j0j0V0j0j0j0C0j0b080b080V0b0j080b0j0j0j0j0j0j0j0b0r0C0j0b0j0j0j0C0j080b0j0j0j0j0j0j0u0r0C0u0j0j0j0j0j0j0b080j0C0j0b080b080b0j0C0j080b0j0j0j0j0j0j0b080b110b0j0j0j0j0j0j0j0j0j0b0r0j0j0j0b0j0j0j0r0b0j0b080j0j0j0j0b080b080b080b0r0b0j080b080b0j0j0j0j0j0j0b0r0C0b080b0j0j0j0j080b0j0b080j0j0j0j0b080b080b0j0j0j0r0b0j0j0j0j0j0j0b080b0j080C0b0j080b080b0j0j0j0j080b0j0b0r0C0b080b0j0j0j0j080b0j0j0j0j0j0b080b080b080b0j0j080b0r0b0j0j0j0j0j0j0b0j0j080C0b0j080b080b0j0j0j0j0j0b080C0j0j0b080b0j0j0C0j0b080j0j0j0j0b080b080b080b0C0C080b0j0j0j0j0j0j0b0C0C080b080b080b0j0j0j0j0j0j0b0C080j0j0b0j0j0j0C0j0b080j0b080j0j0b080b080b080b0C0r0b0j0j0j0j0j0j0b080b0r0b0r0b0j080b080b0j0j0j0j0j0j0b0r0C0j0b0j0j0j0j0j0j0b080j0C0j0b080j0b0j0j0K0b0j0C0j0j0j0b080b0j0K0b0j080b0j0j0j0j0j0j0V0j0j0b0j0j0j0C0j0j0j0j',
            '0l0C0K0N0r0N0j0r1G0V0m0j0V1c0C0j0j0j0j1O0N110u0j0j0j0C0j0j0V0C0j0u110u0j0j0j0C0j0j0j0C0C0j250j1c2S1v1v0j5x2g0j1c0j0j1c2z0j1c0j0j1c0j0N1v0V0C1v0C0b0C0V0j0j0C0j0C1v0u0j0C0C0j0j0j0C0j0j0j0u110u0j0j0j0C0j0C0C0C0b080b0j0C0j080b0j0C0j0j0j0u110u0j0j0j0C0j0j0j0C0j0j0j0u0C0r0u0j0j0j0j0j0j0b0r0b0V080b080b0j0C0j0j0j0V0j0j0b0j0j0j0C0j0j0j0j0j0j0j0b080j0b0C0r0j0b0j0j0j0C0j0b0r0b0r0j0b080b080b0j0C0j0j0j0j0j0j0j0j0b0j0C0r0b0j0j0j0j0j0j0b080b080j0b0r0b0r0j0b0j0j0j0j080b0j0b0r0j0j0j0b080b080b0j0j0j0j080b0j0j0j0j0j0j0b0j0j0j0r0b0j0j0j0j0j0j0b080b080b080b0r0C0b080b0j0j0j0j0j0b080b0r0C0b080b080b080b0j0j0j0j080b0j0C0j0j0j0b0j0j0C080b0j0j0j0j0j0j0b080j0b0C080j0j0b0j0j0j0j0j0j0b0r0b080j0j0b080b080b0j0j0j0j0j0j0b080j0j0j0j0b0j0j0j0r0b0j0b080j0j0j0j0j0b080b080b0C0r0b0j0j0j0j0j0j0b080b080j0C0b0j080b080b0j0j0j0j0j0j',
            '0a0j0j0j0j0C0j0j0C0j0C0C0j0j0j0j0j0j0j0m0C0j0j0j0j0u080j0j0j1n0j0j0j0j0C0j0j0j0V0j0j0j1c0u0j0C0V0j0j0V0j0j1v0N0C0V2o1v1O2S2o141v0j1v4l0j1c0j1v2S2o0C0u1v0j0C0C2S1v0j1c0j0j1v0N251c0j1v0b1c1v1n1v0j0j0V0j0j1v0N1v0C0V0j0j1v0b0C0j0j0V1c0j0u0j1c0j0j0j0j0j0j0j0j1c0j0u0j0j0V0j0j0j0j0j0j0b080u110u0j0j0j0j0j0j1c0j0b0j080b0j0C0j0j0j0V0j0j0u0C0V0j0j0j0C0j0b080j1c0j0b0j0j0j0C0j0C0j0j0j0b080b080b0j0C0j080b0j0j0j0j0j0j0j0b0C0r0u0j0j0j0j0j0j0b080j0b0r0C0j0b0j0j0j0r0b0j0b0r0j0j0j0b080b080b0j0r0b0j080b0j0j0j0j0j0j0b0j0r0C0b0j0j0j0j0j0j0b080j0j0C0j0j0b080b0j0j0j0j0j0j0j0j0j0j0b080b080b080b0C0j0j080b0j0j0j0j0j0j0b0j0j0C080b0j0j0j0j0j0j0j0j0b0C080j0j0b0j0j0j0j0j',
            '0n0Q0j1c14010q0V1c171k0u0r140V0j0j1c0C0N1O0j0V0j0j0j1c0j0u110u0C0j0C0V0C0j0j0b671v0j1v5Q1O2S2o2S1v4l1v0j1v2S2o0C1Z0j0C0C1O141v0j1c0j2z1O0j0V0j0j1v0b2H390j1c0j0V0C2z0j1c0j1v2g0C0V0j1O0b0j0j0V0C1c0j0u0j1c0j0j0j0j0j0j0j0j1c0N0j0j0V0j0j0C0j0j0b081v0u0j0j0j0C0j1c0N0j0j0C0j0j0j0C0j0j0j0u0C0r0u0j0j0j0C0j0b080j1c0j0b0j0C0C0j0C0C0j0b080b080u0C0j080b0j0C0j0j0j0u110u0j0j0j0j0j0j0j0j0C0C0j0b0j0j0j0C0j0C0C0j0b080b080b0j0C0j080b0j0C0j0j0j0b0j110b0j0j0j0j0j',
            '0B0j0V0j0j0C0j0j0j0C0j0C0j0j0C0j0m0j0j0j0j0C0j0C0j0j0u0j1c0j0j0C0C0j0j0j0j0j0j0j0j0u110N0j0j0V0C0V0j0b081n080b0CrU1O5e2SbX2_1Z0V2o141v0j0C0C0j2z1v0j1c0j7N1O420j1c0j1v2S1c0j1v2S2_0b0j0V0j0j1v0N1v0j0j1c0j1v140j0V0j0j0C0C0b080u1v0C0V0u110u0j0j0j0C0j0j0j0C0C0N0C0V0j0j0C0j0j0b080u110u0C0j0C0u0r0C0u080b0j0j0C0j0j0j'
        ]
        for m in months:
            n: int = 0
            size: int = int(m.__len__() / 2)
            l: List[int] = []
            for y in range(0, size):
                z: int = y * 2
                t: int = 0
                c: int = 1
                for x in range(1, -1, -1):
                    t += c * chars.index(m[z + x])
                    c *= 64
                n += t
                l.append(n)
            LunarYear._LEAP.append(l)
        LunarYear._isInit = True

    @staticmethod
    def validate(year: int) -> None:
        if year < -1 or year > 9999:
            raise ValueError(f'illegal lunar year: {year}')

    def __init__(self, year: int):
        LunarYear._init()
        LunarYear.validate(year)
        super().__init__(year)

    @classmethod
    def from_year(cls, year: int) -> LunarYear:
        return cls(year)

    def get_day_count(self) -> int:
        """
        当年的总天数
        :return: 返回为数字，从正月初一到除夕的总天数。
        """
        n: int = 0
        months: List[LunarMonth] = self.get_months()
        for i in range(0, months.__len__()):
            n += months[i].get_day_count()
        return n

    def get_month_count(self) -> int:
        """
        当年的总月数
        :return: 返回为数字，有闰月为13，无闰月为12
        """
        return 12 if self.get_leap_month() < 1 else 13

    def get_name(self) -> str:
        """
        :return: 农历{六十甲子}年
        """
        return f'农历{self.get_sixty_cycle().get_name()}年'

    def next(self, n: int) -> LunarYear:
        return LunarYear(self._year + n)

    def get_leap_month(self) -> int:
        """
        当年的闰月月份
        :return: 返回为数字，代表当年的闰月月份，例如：5代表闰五月，0代表当年没有闰月。
        """
        if self._year == -1:
            return 11
        for i in range(0, self._LEAP.__len__()):
            if self._year in self._LEAP[i]:
                return i + 1
        return 0

    def get_sixty_cycle(self) -> SixtyCycle:
        """
        当年的干支
        :return: 干支 SixtyCycle
        """
        from tyme4py.sixtycycle import SixtyCycle
        return SixtyCycle(self._year - 4)

    def get_twenty(self) -> Twenty:
        """
        :return: 返回为运 Twenty。
        """
        return Twenty(floor((self._year - 1864) / 20))

    def get_nine_star(self) -> NineStar:
        """
        :return: 返回为九星 NineStar。
        """
        return NineStar(63 + self.get_twenty().get_sixty().get_index() * 3 - self.get_sixty_cycle().get_index())

    def get_jupiter_direction(self) -> Direction:
        """
        太岁方位
        :return: 返回为方位 Direction。
        """
        return Direction([0, 7, 7, 2, 3, 3, 8, 1, 1, 6, 0, 0][self.get_sixty_cycle().get_earth_branch().get_index()])

    def get_first_month(self) -> LunarMonth:
        """
        首月（农历月，即一月，俗称正月）
        :return: 农历月
        """
        return LunarMonth.from_ym(self._year, 1)

    def get_months(self) -> List[LunarMonth]:
        """
        农历月列表
        :return: 农历月 LunarMonth 的列表，从正月到十二月，包含闰月。
        """
        l: List[LunarMonth] = []
        m: LunarMonth = self.get_first_month()
        while m.get_year() == self._year:
            l.append(m)
            m = m.next(1)
        return l

    def get_kitchen_god_steed(self) -> KitchenGodSteed:
        """
        灶马头
        :return: 灶马头
        """
        return KitchenGodSteed.from_lunar_year(self._year)


class LunarSeason(LoopTyme):
    """
    农历季节
    从正月开始，依次为：孟春、仲春、季春、孟夏、仲夏、季夏、孟秋、仲秋、季秋、孟冬、仲冬、季冬。
    """
    NAMES: List[str] = ['孟春', '仲春', '季春', '孟夏', '仲夏', '季夏', '孟秋', '仲秋', '季秋', '孟冬', '仲冬', '季冬']
    """名称"""

    def __init__(self, index_or_name: Union[int, str]):
        super().__init__(self.NAMES, index_or_name)

    @classmethod
    def from_name(cls, name: str) -> LunarSeason:
        return cls(name)

    @classmethod
    def from_index(cls, index: int) -> LunarSeason:
        return cls(index)

    def next(self, n: int) -> LunarSeason:
        return LunarSeason(self.next_index(n))


class LunarMonth(MonthUnit):
    """
    农历月
    农历月以初一开始，大月30天，小月29天。
    """
    NAMES: List[str] = ['正月', '二月', '三月', '四月', '五月', '六月', '七月', '八月', '九月', '十月', '十一月', '十二月']
    _leap: bool
    """是否闰月"""

    @staticmethod
    def validate(year: int, month: int) -> None:
        if month == 0 or month > 12 or month < -12:
            raise ValueError(f'illegal lunar month: {month}')
        if month < 0 and -month != LunarYear.from_year(year).get_leap_month():
            raise ValueError(f'illegal leap month {-month} in lunar year {year}')

    def __init__(self, year: int, month: int):
        """
        :param year: 农历年
        :param month: 农历月，闰月为负
        """
        LunarMonth.validate(year, month)
        super().__init__(year, abs(month))
        self._leap = month < 0

    @classmethod
    def from_ym(cls, year: int, month: int) -> LunarMonth:
        """
        从农历年、月初始化
        :param year: 农历年，支持从-1到9999年；
        :param month:  农历月，支持1到12，如果为闰月的，使用负数，即-3代表闰三月。
        :return: LunarMonth
        """
        return cls(year, month)

    def _get_new_moon(self):
        from tyme4py.solar import SolarTerm
        # 冬至
        dong_zhi_jd: float = SolarTerm(self._year, 0).get_cursory_julian_day()
        # 冬至前的初一，今年首朔的日月黄经差
        w: int = ShouXingUtil.calc_shuo(dong_zhi_jd)
        if w > dong_zhi_jd:
            w -= 29.53

        # 正常情况正月初一为第3个朔日，但有些特殊的
        offset: int = 2
        if 8 < self._year < 24:
            offset = 1
        elif LunarYear(self._year - 1).get_leap_month() > 10 and self._year != 239 and self._year != 240:
            offset = 3
        # 本月初一
        return w + 29.5306 * (offset + self.get_index_in_year())

    def get_lunar_year(self) -> LunarYear:
        """
        农历年
        :return: 农历年
        """
        return LunarYear.from_year(self._year)

    def get_month_with_leap(self) -> int:
        """
        月(支持闰月)
        :return: 返回为月份数字，范围为1到12，闰月为负数，如闰7月返回-7。
        """
        return -self._month if self._leap else self._month

    def get_day_count(self) -> int:
        """
        当月的总天数
        :return:返回为数字，从初一开始的总天数，大月有30天，小月有29天。
        """
        w: float = self._get_new_moon()
        return ~~(ShouXingUtil.calc_shuo(w + 29.5306) - ShouXingUtil.calc_shuo(w))

    def get_index_in_year(self) -> int:
        """
        位于当年的月索引
        :return: 返回为数字，范围0到12，正月为0，依次类推，例如五月索引值为4，闰五月索引值为5。
        """
        # 位于当年的索引
        index: int = self._month - 1
        if self.is_leap():
            index += 1
        else:
            leap_month: int = LunarYear.from_year(self._year).get_leap_month()
            if 0 < leap_month < self._month:
                index += 1
        return index

    def get_season(self) -> LunarSeason:
        """
        :return: 农历季节
        """
        return LunarSeason(self._month - 1)

    def get_first_julian_day(self) -> JulianDay:
        """
        初一的儒略日
        :return: 儒略日
        """
        from tyme4py.jd import JulianDay
        return JulianDay.from_julian_day(JulianDay.J2000 + ShouXingUtil.calc_shuo(self._get_new_moon()))

    def is_leap(self) -> bool:
        """
        是否闰月
        :return: 返回为布尔值，闰月返回true，非闰月返回false。
        """
        return self._leap

    def get_week_count(self, start: int) -> int:
        """
        当月有几周
        :param start: 参数为起始星期，1234560分别代表星期一至星期天
        :return: 返回为数字。
        """
        return ceil((self.index_of(self.get_first_julian_day().get_week().get_index() - start, 7) + self.get_day_count()) / 7)

    def get_name(self) -> str:
        """
        依据国家标准《农历的编算和颁行》GB/T 33661-2017中农历月的命名方法。
        :return: 名称
        """
        return ('闰' if self._leap else '') + self.NAMES[self._month - 1]

    def __str__(self) -> str:
        return f'{self.get_lunar_year()}{self.get_name()}'

    def next(self, n: int) -> LunarMonth:
        if n == 0:
            return LunarMonth.from_ym(self._year, self.get_month_with_leap())

        m: int = self.get_index_in_year() + 1 + n
        y: LunarYear = self.get_lunar_year()
        if n > 0:
            month_count: int = y.get_month_count()
            while m > month_count:
                m -= month_count
                y = y.next(1)
                month_count = y.get_month_count()
        else:
            while m <= 0:
                y = y.next(-1)
                m += y.get_month_count()

        leap: bool = False
        leap_month: int = y.get_leap_month()
        if leap_month > 0:
            if m == leap_month + 1:
                leap = True
            if m > leap_month:
                m -= 1
        return LunarMonth.from_ym(y.get_year(), -m if leap else m)

    def get_days(self) -> List[LunarDay]:
        """
        本月的农历日列表
        :return: 农历日 LunarDay 的列表，从初一开始。
        """
        m: int = self.get_month_with_leap()
        l: List[LunarDay] = []
        for i in range(1, self.get_day_count() + 1):
            l.append(LunarDay(self._year, m, i))
        return l

    def get_first_day(self) -> LunarDay:
        """
        初一
        """
        return LunarDay.from_ymd(self._year, self.get_month_with_leap(), 1)

    def get_weeks(self, start: int) -> List[LunarWeek]:
        """
        本月的农历周列表
        :param start: 参数为起始星期，1234560分别代表星期一至星期天
        :return: 农历周 LunarWeek 的列表。
        """
        m: int = self.get_month_with_leap()
        l: List[LunarWeek] = []
        for i in range(0, self.get_week_count(start)):
            l.append(LunarWeek(self._year, m, i, start))
        return l

    def get_sixty_cycle(self) -> SixtyCycle:
        """
        当月的干支
        :return: 干支 SixtyCycle。
        """
        from tyme4py.sixtycycle import SixtyCycle, HeavenStem, EarthBranch
        return SixtyCycle(HeavenStem(self.get_lunar_year().get_sixty_cycle().get_heaven_stem().get_index() * 2 + self._month + 1).get_name() + EarthBranch(self._month + 1).get_name())

    def get_nine_star(self) -> NineStar:
        """
        :return: 九星 NineStar。
        """
        index = self.get_sixty_cycle().get_earth_branch().get_index()
        if index < 2:
            index += 3
        return NineStar(27 - self.get_lunar_year().get_sixty_cycle().get_earth_branch().get_index() % 3 * 3 - index)

    def get_jupiter_direction(self) -> Direction:
        """
        太岁方位
        :return: 方位 Direction。
        """
        from tyme4py.sixtycycle import SixtyCycle
        sixty_cycle: SixtyCycle = self.get_sixty_cycle()
        n: int = [7, -1, 1, 3][sixty_cycle.get_earth_branch().next(-2).get_index() % 4]
        return Direction(n) if n != -1 else sixty_cycle.get_heaven_stem().get_direction()

    def get_fetus(self) -> Union[FetusMonth, None]:
        """
        :return: 返回为逐月胎神 FetusMonth。闰月无胎神。
        """
        return FetusMonth.from_lunar_month(self)

    def get_minor_ren(self) -> MinorRen:
        """
        :return: 小六壬
        """
        return MinorRen((self._month - 1) % 6)


class LunarWeek(WeekUnit):
    """
    农历周
    农历一个月最多有6个周，分别为：第一周、第二周、第三周、第四周、第五周、第六周。
    """
    NAMES: List[str] = ['第一周', '第二周', '第三周', '第四周', '第五周', '第六周']
    """名称"""

    @staticmethod
    def validate(year: int, month: int, index: int, start: int) -> None:
        WeekUnit.validate(year, month, index, start)
        m: LunarMonth = LunarMonth.from_ym(year, month)
        if index >= m.get_week_count(start):
            raise ValueError(f'illegal lunar week index: {index} in month: {m}')

    def __init__(self, year: int, month: int, index: int, start: int):
        """
        :param year: 年
        :param month: 月
        :param index: 周索引，0-5
        :param start: 起始星期  (1234560分别代表星期一至星期日)
        """
        LunarWeek.validate(year, month, index, start)
        super().__init__(year, month, index, start)

    @classmethod
    def from_ym(cls, year: int, month: int, index: int, start: int) -> LunarWeek:
        return cls(year, month, index, start)

    def get_lunar_month(self) -> LunarMonth:
        """
        :return:农历月
        """
        return LunarMonth.from_ym(self._year, self._month)

    def get_name(self) -> str:
        return self.NAMES[self._index]

    def __str__(self) -> str:
        return f'{self.get_lunar_month()}{self.get_name()}'

    def next(self, n: int) -> LunarWeek:
        if n == 0:
            return LunarWeek(self._year, self._month, self._index, self._start)

        d: int = self._index + n
        m: LunarMonth = self.get_lunar_month()
        if n > 0:
            week_count: int = m.get_week_count(self._start)
            while d >= week_count:
                d -= week_count
                m = m.next(1)
                if m.get_first_day().get_week().get_index() != self._start:
                    d += 1
                week_count = m.get_week_count(self._start)
        else:
            while d < 0:
                if m.get_first_day().get_week().get_index() != self._start:
                    d -= 1
                m = m.next(-1)
                d += m.get_week_count(self._start)
        return LunarWeek(m.get_year(), m.get_month_with_leap(), d, self._start)

    def get_first_day(self) -> LunarDay:
        """
        本周第一天的农历日
        :return: 农历日 LunarDay。
        """
        first_day: LunarDay = LunarDay(self._year, self._month, 1)
        return first_day.next(self._index * 7 - self.index_of(first_day.get_week().get_index() - self._start, 7))

    def get_days(self) -> List[LunarDay]:
        """
        本周农历日列表
        :return: 农历日 LunarDay的列表。
        """
        l: List[LunarDay] = []
        d: LunarDay = self.get_first_day()
        l.append(d)
        for i in range(1, 7):
            l.append(d.next(i))
        return l

    def __eq__(self, other: LunarWeek) -> bool:
        return other and other.get_first_day() == self.get_first_day()


class LunarDay(DayUnit):
    """农历日"""
    NAMES: List[str] = ['初一', '初二', '初三', '初四', '初五', '初六', '初七', '初八', '初九', '初十', '十一', '十二', '十三', '十四', '十五', '十六', '十七', '十八', '十九', '二十', '廿一', '廿二', '廿三', '廿四', '廿五', '廿六', '廿七', '廿八', '廿九', '三十']
    """名称"""

    @staticmethod
    def validate(year: int, month: int, day: int) -> None:
        if day < 1:
            raise ValueError(f'illegal lunar day {day}')
        m: LunarMonth = LunarMonth.from_ym(year, month)
        if day > m.get_day_count():
            raise ValueError(f'illegal day {day} in {m}')

    def __init__(self, year: int, month: int, day: int):
        """
        :param year: 农历年
        :param month: 农历月，闰月为负
        :param day: 农历日
        """
        LunarDay.validate(year, month, day)
        super().__init__(year, month, day)

    @classmethod
    def from_ymd(cls, year: int, month: int, day: int) -> LunarDay:
        return cls(year, month, day)

    def get_lunar_month(self) -> LunarMonth:
        """
        :return: 农历月
        """
        return LunarMonth.from_ym(self._year, self._month)

    def get_name(self) -> str:
        """
        :return:农历日名，'初一'
        """
        return self.NAMES[self._day - 1]

    def __str__(self) -> str:
        return f'{self.get_lunar_month()}{self.get_name()}'

    def next(self, n: int) -> LunarDay:
        return self.get_solar_day().next(n).get_lunar_day()

    def is_before(self, target: LunarDay) -> bool:
        """
        是否在指定农历日之前
        :param target:农历日 LunarDay
        :return: true/false
        """
        y: int = target.get_year()
        if self._year != y:
            return self._year < y
        m: int = target.get_month()
        if self._month != m:
            return abs(self._month) < abs(m)
        return self._day < target.get_day()

    def is_after(self, target: LunarDay) -> bool:
        """
        是否在指定农历日之后
        :param target:农历日 LunarDay
        :return:
        """
        y: int = target.get_year()
        if self._year != y:
            return self._year > y
        m: int = target.get_month()
        if self._month != m:
            return abs(self._month) >= abs(m)
        return self._day > target.get_day()

    def get_week(self) -> Week:
        """
        :return: 星期
        """
        return self.get_solar_day().get_week()

    def get_year_sixty_cycle(self) -> SixtyCycle:
        """
        当天的年干支 非当天所属的农历年干支，以立春换年。
        :return: 干支 SixtyCycle
        """
        warnings.warn('get_year_sixty_cycle() is deprecated, please use SixtyCycleDay.get_year() instead.', DeprecationWarning)
        return self.get_sixty_cycle_day().get_year()

    def get_month_sixty_cycle(self) -> SixtyCycle:
        """
        当天的月干支 非当天所属的农历月干支，以节令换月。
        :return: 干支 SixtyCycle
        """
        warnings.warn('get_month_sixty_cycle() is deprecated, please use SixtyCycleDay.get_month() instead.', DeprecationWarning)
        return self.get_sixty_cycle_day().get_month()

    def get_sixty_cycle(self) -> SixtyCycle:
        """
        当天的干支
        :return: 干支
        """
        from tyme4py.sixtycycle import SixtyCycle, HeavenStem, EarthBranch
        offset: int = int(self.get_lunar_month().get_first_julian_day().next(self._day - 12).get_day())
        return SixtyCycle(HeavenStem(offset).get_name() + EarthBranch(offset).get_name())

    def get_duty(self) -> Duty:
        """
        建除十二值神
        :return: 建除十二值神 Duty
        """
        return self.get_sixty_cycle_day().get_duty()

    def get_twelve_star(self) -> TwelveStar:
        """
        黄道黑道十二神
        :return: 黄道黑道十二神 TwelveStar
        """
        return self.get_sixty_cycle_day().get_twelve_star()

    def get_nine_star(self) -> NineStar:
        """
        :return: 九星 NineStar。
        """
        from tyme4py.solar import SolarTerm
        d: SolarDay = self.get_solar_day()
        dong_zhi: SolarTerm = SolarTerm(d.get_year(), 0)
        dong_zhi_solar: SolarDay = dong_zhi.get_solar_day()
        xia_zhi_solar: SolarDay = dong_zhi.next(12).get_solar_day()
        dong_zhi_solar2: SolarDay = dong_zhi.next(24).get_solar_day()
        dong_zhi_index: int = dong_zhi_solar.get_lunar_day().get_sixty_cycle().get_index()
        xia_zhi_index: int = xia_zhi_solar.get_lunar_day().get_sixty_cycle().get_index()
        dong_zhi_index2: int = dong_zhi_solar2.get_lunar_day().get_sixty_cycle().get_index()
        solar_shun_bai: SolarDay = dong_zhi_solar.next(60 - dong_zhi_index if dong_zhi_index > 29 else -dong_zhi_index)
        solar_shun_bai2: SolarDay = dong_zhi_solar2.next(60 - dong_zhi_index2 if dong_zhi_index2 > 29 else -dong_zhi_index2)
        solar_ni_zi: SolarDay = xia_zhi_solar.next(60 - xia_zhi_index if xia_zhi_index > 29 else -xia_zhi_index)
        offset: int = 0
        if not d.is_before(solar_shun_bai) and d.is_before(solar_ni_zi):
            offset = d.subtract(solar_shun_bai)
        elif not d.is_before(solar_ni_zi) and d.is_before(solar_shun_bai2):
            offset = 8 - d.subtract(solar_ni_zi)
        elif not d.is_before(solar_shun_bai2):
            offset = d.subtract(solar_shun_bai2)
        elif d.is_before(solar_shun_bai):
            offset = 8 + solar_shun_bai.subtract(d)
        return NineStar(offset)

    def get_jupiter_direction(self) -> Direction:
        """
        太岁方位
        :return: 方位 Direction。
        """
        index: int = self.get_sixty_cycle().get_index()
        from tyme4py.culture import Element
        return Element(index // 12).get_direction() if index % 12 < 6 else self.get_lunar_month().get_lunar_year().get_jupiter_direction()

    def get_fetus_day(self) -> FetusDay:
        """
        :return:逐日胎神 FetusDay。
        """
        return FetusDay.from_lunar_day(self)

    def get_phase_day(self) -> PhaseDay:
        """
        月相第几天
        :return: 月相第几天
        """
        today = self.get_solar_day()
        m = self.get_lunar_month().next(1)
        p = Phase.from_index(m.get_year(), m.get_month_with_leap(), 0)
        d = p.get_solar_day()
        while d.is_after(today):
            p = p.next(-1)
            d = p.get_solar_day()
        return PhaseDay(p, today.subtract(d))

    def get_phase(self) -> Phase:
        """
        月相
        :return: 月相 Phase。
        """
        return self.get_phase_day().get_phase()

    def get_solar_day(self) -> SolarDay:
        """
        公历日
        :return: 公历日 SolarDay。
        """
        return self.get_lunar_month().get_first_julian_day().next(self._day - 1).get_solar_day()

    def get_sixty_cycle_day(self) -> SixtyCycleDay:
        """
        干支日
        :return: 干支日
        """
        return self.get_solar_day().get_sixty_cycle_day()

    def get_twenty_eight_star(self) -> TwentyEightStar:
        """
        :return: 二十八宿 TwentyEightStar。
        """
        return TwentyEightStar([10, 18, 26, 6, 14, 22, 2][self.get_solar_day().get_week().get_index()]).next(-7 * self.get_sixty_cycle().get_earth_branch().get_index())

    def get_festival(self) -> Union[LunarFestival, None]:
        """
        农历传统节日，如果当天不是农历传统节日，返回None
        :return:农历传统节日 LunarFestival
        """
        return LunarFestival.from_ymd(self._year, self._month, self._day)

    def get_six_star(self) -> SixStar:
        """
        :return: 六曜
        """
        return SixStar((abs(self._month) + self._day - 2) % 6)

    def get_gods(self) -> List[God]:
        """
        神煞列表(吉神宜趋，凶神宜忌)
        :return: 神煞列表
        """
        return self.get_sixty_cycle_day().get_gods()

    def get_recommends(self) -> List[Taboo]:
        """
        今日 宜
        :return: 宜忌列表
        """
        return self.get_sixty_cycle_day().get_recommends()

    def get_avoids(self) -> List[Taboo]:
        """
        今日 忌
        :return: 宜忌列表
        """
        return self.get_sixty_cycle_day().get_avoids()

    def get_hours(self) -> List[LunarHour]:
        """
        当天的时辰列表
        :return: 由于23:00-23:59、00:00-00:59均为子时，而农历日是从00:00-23:59为一天，所以获取当天的时辰列表，实际会返回13个。
        """
        l: List[LunarHour] = [LunarHour(self._year, self._month, self._day, 0, 0, 0)]
        for i in range(0, 24, 2):
            l.append(LunarHour(self._year, self._month, self._day, i + 1, 0, 0))
        return l

    def get_minor_ren(self) -> MinorRen:
        """
        小六壬
        :return: 小六壬
        """
        return self.get_lunar_month().get_minor_ren().next(self._day - 1)

    def get_three_pillars(self) -> ThreePillars:
        """
        三柱
        @return: 三柱
        """
        return self.get_sixty_cycle_day().get_three_pillars()


class LunarHour(SecondUnit):
    """
    农历时辰
    """
    provider: EightCharProvider = DefaultEightCharProvider()
    """八字计算接口"""

    @staticmethod
    def validate(year: int, month: int, day: int, hour: int, minute: int, second: int) -> None:
        SecondUnit.validate(year, month, day, hour, minute, second)
        LunarDay.validate(year, month, day)

    def __init__(self, year: int, month: int, day: int, hour: int, minute: int = 0, second: int = 0):
        """
        :param year: 农历年，支持从-1到9999年；
        :param month: 农历月，支持1到12，如果为闰月的，使用负数，即-3代表闰三月；
        :param day: 农历日，支持1到30，大月30天，小月29天。
        :param hour: 时为0-23
        :param minute: 分为0-59
        :param second: 秒为0-59
        """
        LunarHour.validate(year, month, day, hour, minute, second)
        super().__init__(year, month, day, hour, minute, second)

    @classmethod
    def from_ymd_hms(cls, year: int, month: int, day: int, hour: int, minute: int = 0, second: int = 0) -> LunarHour:
        return cls(year, month, day, hour, minute, second)

    def get_lunar_day(self) -> LunarDay:
        """
        :return: 农历日
        """
        return LunarDay.from_ymd(self._year, self._month, self._day)

    def get_name(self) -> str:
        from tyme4py.sixtycycle import EarthBranch
        return f'{EarthBranch(self.get_index_in_day()).get_name()}时'

    def __str__(self) -> str:
        return f'{self.get_lunar_day()}{self.get_sixty_cycle().get_name()}时'

    def get_index_in_day(self) -> int:
        """
        位于当天的索引
        :return: 数字，范围0到11。
        """
        return int((self._hour + 1) / 2)

    def next(self, n: int) -> LunarHour:
        if n == 0:
            return LunarHour(self._year, self._month, self._day, self._hour, self._minute, self._second)
        h: int = self._hour + n * 2
        diff: int = -1 if h < 0 else 1
        hour: int = abs(h)
        days: int = int(hour / 24) * diff
        hour = (hour % 24) * diff
        if hour < 0:
            hour += 24
            days -= 1
        d: LunarDay = self.get_lunar_day().next(days)
        return LunarHour(d.get_year(), d.get_month(), d.get_day(), hour, self._minute, self._second)

    def is_before(self, target: LunarHour) -> bool:
        """
        是否在指定农历时辰之前
        :param target: 农历时辰
        :return: true/false
        """
        a_day: LunarDay = self.get_lunar_day()
        b_day: LunarDay = target.get_lunar_day()
        if a_day != b_day:
            return a_day.is_before(b_day)
        h = target.get_hour()
        if self._hour != h:
            return self._hour < h
        m = target.get_minute()
        return self._minute < m if self._minute != m else self._second < target.get_second()

    def is_after(self, target: LunarHour) -> bool:
        """
        是否在指定农历时辰之后
        :param target: 农历时辰
        :return: true/false
        """
        a_day: LunarDay = self.get_lunar_day()
        b_day: LunarDay = target.get_lunar_day()
        if a_day != b_day:
            return a_day.is_after(b_day)
        h = target.get_hour()
        if self._hour != h:
            return self._hour > h
        m = target.get_minute()
        return self._minute > m if self._minute != m else self._second > target.get_second()

    def get_year_sixty_cycle(self) -> SixtyCycle:
        """
        当时的年干支  非当时所属的农历年干支
        :return: 干支 SixtyCycle
        """
        warnings.warn('get_year_sixty_cycle() is deprecated, please use SixtyCycleHour.get_year() instead.', DeprecationWarning)
        return self.get_sixty_cycle_hour().get_year()

    def get_month_sixty_cycle(self) -> SixtyCycle:
        """
        当时的月干支 非当天所属的农历月干支，以节令具体时刻换月。
        :return: 干支 SixtyCycle
        """
        warnings.warn('get_year_sixty_cycle() is deprecated, please use SixtyCycleHour.get_month() instead.', DeprecationWarning)
        return self.get_sixty_cycle_hour().get_month()

    def get_day_sixty_cycle(self) -> SixtyCycle:
        """
        当时的日干支
        :return: 干支 SixtyCycle。注意：23:00开始算做第二天。
        """
        warnings.warn('get_year_sixty_cycle() is deprecated, please use SixtyCycleHour.get_day() instead.', DeprecationWarning)
        return self.get_sixty_cycle_hour().get_day()

    def get_sixty_cycle(self) -> SixtyCycle:
        """
        时辰干支
        :return: 干支 SixtyCycle。
        """
        from tyme4py.sixtycycle import SixtyCycle, HeavenStem, EarthBranch
        earth_branch_index: int = self.get_index_in_day() % 12
        d: SixtyCycle = self.get_lunar_day().get_sixty_cycle()
        if self._hour >= 23:
            d = d.next(1)
        return SixtyCycle(HeavenStem(d.get_heaven_stem().get_index() % 5 * 2 + earth_branch_index).get_name() + EarthBranch(earth_branch_index).get_name())

    def get_twelve_star(self) -> TwelveStar:
        """
        :return: 黄道黑道十二神 TwelveStar。
        """
        return TwelveStar(self.get_sixty_cycle().get_earth_branch().get_index() + (8 - self.get_sixty_cycle_hour().get_day().get_earth_branch().get_index() % 6) * 2)

    def get_nine_star(self) -> NineStar:
        """
        九星（时家紫白星歌诀：三元时白最为佳，冬至阳生顺莫差，孟日七宫仲一白，季日四绿发萌芽，每把时辰起甲子，本时星耀照光华，时星移入中宫去，顺飞八方逐细查。夏至阴生逆回首，孟归三碧季加六，仲在九宫时起甲，依然掌中逆轮跨。）
        :return: 九星 NineStar。
        """
        from tyme4py.solar import SolarTerm
        d: LunarDay = self.get_lunar_day()
        solar: SolarDay = d.get_solar_day()
        dong_zhi: SolarTerm = SolarTerm(solar.get_year(), 0)
        earth_branch_index: int = self.get_index_in_day() % 12
        index: int = [8, 5, 2][d.get_sixty_cycle().get_earth_branch().get_index() % 3]
        if (not solar.is_before(dong_zhi.get_julian_day().get_solar_day())) and solar.is_before(dong_zhi.next(12).get_julian_day().get_solar_day()):
            index = 8 + earth_branch_index - index
        else:
            index -= earth_branch_index
        return NineStar(index)

    def get_solar_time(self) -> SolarTime:
        """
        公历时刻
        :return: 公历时刻 SolarTime
        """
        from tyme4py.solar import SolarTime
        d: SolarDay = self.get_lunar_day().get_solar_day()
        return SolarTime(d.get_year(), d.get_month(), d.get_day(), self._hour, self._minute, self._second)

    def get_sixty_cycle_hour(self) -> SixtyCycleHour:
        """
        干支时辰
        :return: 干支时辰 SixtyCycleHour
        """
        return self.get_solar_time().get_sixty_cycle_hour()

    def get_eight_char(self) -> EightChar:
        """
        八字
        :return: 八字
        """
        return LunarHour.provider.get_eight_char(self)

    def get_recommends(self) -> List[Taboo]:
        """
        宜
        :return: 宜忌列表
        """
        return Taboo.get_hour_recommends(self.get_sixty_cycle_hour().get_day(), self.get_sixty_cycle())

    def get_avoids(self) -> List[Taboo]:
        """
        忌
        :return: 宜忌列表
        """
        return Taboo.get_hour_avoids(self.get_sixty_cycle_hour().get_day(), self.get_sixty_cycle())

    def get_minor_ren(self) -> MinorRen:
        """
        :return: 小六壬
        """
        return self.get_lunar_day().get_minor_ren().next(self.get_index_in_day())



# -*- coding:utf-8 -*-
import unittest

from tyme4py.solar import SolarDay


class TestSixStar(unittest.TestCase):
    def test0(self):
        assert SolarDay(2020, 4, 23).get_lunar_day().get_six_star().get_name(), '佛灭'

    def test1(self):
        assert SolarDay(2021, 1, 15).get_lunar_day().get_six_star().get_name() == '友引'

    def test2(self):
        assert SolarDay(2017, 1, 5).get_lunar_day().get_six_star().get_name() == '先胜'

    def test3(self):
        assert SolarDay(2020, 4, 10).get_lunar_day().get_six_star().get_name() == '友引'

    def test4(self):
        assert SolarDay(2020, 6, 11).get_lunar_day().get_six_star().get_name() == '大安'

    def test5(self):
        assert SolarDay(2020, 6, 1).get_lunar_day().get_six_star().get_name() == '先胜'

    def test6(self):
        assert SolarDay(2020, 12, 8).get_lunar_day().get_six_star().get_name() == '先负'

    def test8(self):
        assert SolarDay(2020, 12, 11).get_lunar_day().get_six_star().get_name() == '赤口'



# -*- coding:utf-8 -*-
import unittest

from tyme4py.solar import SolarMonth


class TestSolarMonth(unittest.TestCase):
    def test0(self):
        m = SolarMonth(2019, 5)
        assert m.get_name() == '5月'
        assert m.__str__() == '2019年5月'

    def test1(self):
        m = SolarMonth(2023, 1)
        assert m.get_week_count(0) == 5
        assert m.get_week_count(1) == 6
        assert m.get_week_count(2) == 6
        assert m.get_week_count(3) == 5
        assert m.get_week_count(4) == 5
        assert m.get_week_count(5) == 5
        assert m.get_week_count(6) == 5

    def test2(self):
        m = SolarMonth(2023, 2)
        assert m.get_week_count(0) == 5
        assert m.get_week_count(1) == 5
        assert m.get_week_count(2) == 5
        assert m.get_week_count(3) == 4
        assert m.get_week_count(4) == 5
        assert m.get_week_count(5) == 5
        assert m.get_week_count(6) == 5

    def test3(self):
        m = SolarMonth(2023, 10).next(1)
        assert m.get_name() == '11月'
        assert m.__str__() == '2023年11月'

    def test4(self):
        m = SolarMonth(2023, 10)
        assert m.next(2).__str__() == '2023年12月'
        assert m.next(3).__str__() == '2024年1月'
        assert m.next(-5).__str__() == '2023年5月'
        assert m.next(-9).__str__() == '2023年1月'
        assert m.next(-10).__str__() == '2022年12月'
        assert m.next(24).__str__() == '2025年10月'
        assert m.next(-24).__str__() == '2021年10月'



# -*- coding:utf-8 -*-
import unittest

from tyme4py.solar import SolarDay


class TestJulianDay(unittest.TestCase):
    def test(self):
        assert SolarDay(2023, 1, 1).get_julian_day().get_solar_day().__str__() == '2023年1月1日'



# -*- coding:utf-8 -*-
import unittest

from tyme4py.solar import SolarDay


class TestNineDay(unittest.TestCase):
    def test(self):
        d = SolarDay(2020, 12, 21).get_nine_day()
        assert d
        assert d.get_name() == '一九'
        assert d.get_nine().__str__() == '一九'
        assert d.__str__() == '一九第1天'

    def test1(self):
        d = SolarDay(2020, 12, 22).get_nine_day()
        assert d
        assert d.get_name() == '一九'
        assert d.get_nine().__str__() == '一九'
        assert d.__str__() == '一九第2天'

    def test2(self):
        d = SolarDay(2020, 1, 7).get_nine_day()
        assert d
        assert d.get_name() == '二九'
        assert d.get_nine().__str__() == '二九'
        assert d.__str__() == '二九第8天'

    def test3(self):
        d = SolarDay(2021, 1, 6).get_nine_day()
        assert d
        assert d.get_name() == '二九'
        assert d.get_nine().__str__() == '二九'
        assert d.__str__() == '二九第8天'

    def test4(self):
        d = SolarDay(2021, 1, 8).get_nine_day()
        assert d
        assert d.get_name() == '三九'
        assert d.get_nine().__str__() == '三九'
        assert d.__str__() == '三九第1天'

    def test5(self):
        d = SolarDay(2021, 3, 5).get_nine_day()
        assert d
        assert d.get_name() == '九九'
        assert d.get_nine().__str__() == '九九'
        assert d.__str__() == '九九第3天'

    def test6(self):
        d = SolarDay(2021, 7, 5).get_nine_day()
        assert d is None



# -*- coding:utf-8 -*-
from __future__ import annotations

from typing import List, Union

from tyme4py import LoopTyme


class SixStar(LoopTyme):
    """六曜"""
    NAMES: List[str] = ['先胜', '友引', '先负', '佛灭', '大安', '赤口']
    """名称"""

    def __init__(self, index_or_name: Union[int, str]):
        super().__init__(self.NAMES, index_or_name)

    @classmethod
    def from_name(cls, name: str) -> SixStar:
        return cls(name)

    @classmethod
    def from_index(cls, index: int) -> SixStar:
        return cls(index)

    def next(self, n: int) -> SixStar:
        return SixStar(self.next_index(n))



# -*- coding:utf-8 -*-
from __future__ import annotations
from tyme4py import AbstractTyme
from tyme4py.culture import Week
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from tyme4py.solar import SolarDay, SolarTime


class JulianDay(AbstractTyme):
    """儒略日"""
    J2000: float = 2451545
    """2000年儒略日数(2000-1-1 12:00:00 UTC)"""
    _day: float
    """儒略日"""

    def __init__(self, day: float):
        """
        :param day: 为小数
        """
        self._day = day

    @classmethod
    def from_julian_day(cls, day: float) -> JulianDay:
        return cls(day)

    @classmethod
    def from_ymd_hms(cls, year: int, month: int, day: int, hour: int, minute: int, second: int) -> JulianDay:
        """
        通过公历年月日时分秒进行初始化
        :param year:  年
        :param month: 月
        :param day:   日
        :param hour:  时
        :param minute: 分
        :param second: 秒
        :return:
        """
        d: float = day + ((second / 60 + minute) / 60 + hour) / 24
        n: int = 0
        g: bool = year * 372 + month * 31 + int(d) >= 588829
        if month <= 2:
            month += 12
            year -= 1
        if g:
            n = int(year * 0.01)
            n = 2 - n + int(n * 0.25)
        return cls(int(365.25 * (year + 4716)) + int(30.6001 * (month + 1)) + d + n - 1524.5)

    def get_day(self) -> float:
        """
        :return: 儒略日
        """
        return self._day

    def get_name(self) -> str:
        return f'{self._day}'

    def next(self, n: float) -> JulianDay:
        return JulianDay(self._day + n)

    def get_solar_day(self) -> SolarDay:
        """
        :return: 公历日
        """
        return self.get_solar_time().get_solar_day()

    def get_solar_time(self) -> SolarTime:
        """
        :return: 公历时刻
        """
        d: int = int(self._day + 0.5)
        f: float = self._day + 0.5 - d
        if d >= 2299161:
            c: int = int((d - 1867216.25) / 36524.25)
            d += 1 + c - int(c * 0.25)
        d += 1524
        y: int = int((d - 122.1) / 365.25)
        d -= int(365.25 * y)
        m: int = int(d / 30.601)
        d -= int(30.601 * m)
        if m > 13:
            m -= 12
        else:
            y -= 1
        m -= 1
        y -= 4715
        f *= 24
        hour: int = int(f)
        f -= hour
        f *= 60
        minute: int = int(f)
        f -= minute
        f *= 60
        second: int = round(f)
        from tyme4py.solar import SolarTime
        return SolarTime(y, m, d, hour, minute, second) if second < 60 else SolarTime(y, m, d, hour, minute, second - 60).next(60)

    def get_week(self) -> Week:
        """
        通过儒略日计算的星期是最准的，基姆拉尔森和蔡勒公式计算星期的准确性，在儒略日面前都是弟弟，不服来辩
        :return: 星期
        """
        return Week(int(self._day + 0.5) + 7000001)

    def subtract(self, target: JulianDay) -> float:
        """
        儒略日相减
        :param target: 儒略日
        :return: 差
        """
        return self._day - target.get_day()



# -*- coding:utf-8 -*-
import unittest

from tyme4py.solar import SolarDay


class TestConstellation(unittest.TestCase):
    def test(self):
        assert SolarDay(2020, 3, 21).get_constellation().get_name() == '白羊'
        assert SolarDay(2020, 4, 19).get_constellation().get_name() == '白羊'

    def test1(self):
        assert SolarDay(2020, 4, 20).get_constellation().get_name() == '金牛'
        assert SolarDay(2020, 5, 20).get_constellation().get_name() == '金牛'

    def test2(self):
        assert SolarDay(2020, 5, 21).get_constellation().get_name() == '双子'
        assert SolarDay(2020, 6, 21).get_constellation().get_name() == '双子'

    def test3(self):
        assert SolarDay(2020, 6, 22).get_constellation().get_name() == '巨蟹'
        assert SolarDay(2020, 7, 22).get_constellation().get_name() == '巨蟹'



# -*- coding:utf-8 -*-
import unittest

from tyme4py.solar import SolarDay


class TestDogDay(unittest.TestCase):
    def test(self):
        d = SolarDay(2011, 7, 14).get_dog_day()
        assert d
        assert d.get_name() == '初伏'
        assert d.get_dog().__str__() == '初伏'
        assert d.__str__() == '初伏第1天'

    def test1(self):
        d = SolarDay(2011, 7, 23).get_dog_day()
        assert d
        assert d.get_name() == '初伏'
        assert d.get_dog().__str__() == '初伏'
        assert d.__str__() == '初伏第10天'

    def test2(self):
        d = SolarDay(2011, 7, 24).get_dog_day()
        assert d
        assert d.get_name() == '中伏'
        assert d.get_dog().__str__() == '中伏'
        assert d.__str__() == '中伏第1天'

    def test3(self):
        d = SolarDay(2011, 8, 12).get_dog_day()
        assert d
        assert d.get_name() == '中伏'
        assert d.get_dog().__str__() == '中伏'
        assert d.__str__() == '中伏第20天'

    def test4(self):
        d = SolarDay(2011, 8, 13).get_dog_day()
        assert d
        assert d.get_name() == '末伏'
        assert d.get_dog().__str__() == '末伏'
        assert d.__str__() == '末伏第1天'

    def test5(self):
        d = SolarDay(2011, 8, 22).get_dog_day()
        assert d
        assert d.get_name() == '末伏'
        assert d.get_dog().__str__() == '末伏'
        assert d.__str__() == '末伏第10天'

    def test6(self):
        assert SolarDay(2011, 7, 13).get_dog_day() is None

    def test7(self):
        assert SolarDay(2011, 8, 23).get_dog_day() is None

    def test8(self):
        d = SolarDay(2012, 7, 18).get_dog_day()
        assert d
        assert d.get_name() == '初伏'
        assert d.get_dog().__str__() == '初伏'
        assert d.__str__() == '初伏第1天'

    def test9(self):
        d = SolarDay(2012, 8, 5).get_dog_day()
        assert d
        assert d.get_name() == '中伏'
        assert d.get_dog().__str__() == '中伏'
        assert d.__str__() == '中伏第9天'

    def test10(self):
        d = SolarDay(2012, 8, 8).get_dog_day()
        assert d
        assert d.get_name() == '末伏'
        assert d.get_dog().__str__() == '末伏'
        assert d.__str__() == '末伏第2天'



# -*- coding:utf-8 -*-
import unittest

from tyme4py.solar import SolarDay


class TestSolarDay(unittest.TestCase):
    def test0(self):
        assert SolarDay(2023, 1, 1).get_name() == '1日'
        assert SolarDay(2023, 1, 1).__str__() == '2023年1月1日'

    def test1(self):
        assert SolarDay(2000, 2, 29).get_name() == '29日'
        assert SolarDay(2000, 2, 29).__str__() == '2000年2月29日'

    def test2(self):
        assert SolarDay(2023, 1, 1).get_index_in_year() == 0
        assert SolarDay(2023, 12, 31).get_index_in_year() == 364
        assert SolarDay(2020, 12, 31).get_index_in_year() == 365

    def test3(self):
        assert SolarDay(2023, 1, 1).subtract(SolarDay(2023, 1, 1)) == 0
        assert SolarDay(2023, 1, 2).subtract(SolarDay(2023, 1, 1)) == 1
        assert SolarDay(2023, 1, 1).subtract(SolarDay(2023, 1, 2)) == -1
        assert SolarDay(2023, 2, 1).subtract(SolarDay(2023, 1, 1)) == 31
        assert SolarDay(2023, 1, 1).subtract(SolarDay(2023, 2, 1)) == -31
        assert SolarDay(2024, 1, 1).subtract(SolarDay(2023, 1, 1)) == 365
        assert SolarDay(2023, 1, 1).subtract(SolarDay(2024, 1, 1)) == -365
        assert SolarDay(1582, 10, 15).subtract(SolarDay(1582, 10, 4)) == 1

    def test4(self):
        assert SolarDay(1582, 10, 15).next(-1).__str__() == '1582年10月4日'

    def test5(self):
        assert SolarDay(2000, 2, 28).next(2).__str__() == '2000年3月1日'

    def test6(self):
        assert SolarDay(2020, 5, 24).get_lunar_day().__str__() == '农历庚子年闰四月初二'

    def test7(self):
        assert SolarDay(2020, 5, 24).subtract(SolarDay(2020, 4, 23)) == 31

    def test8(self):
        assert SolarDay(16, 11, 30).get_lunar_day().__str__() == '农历丙子年十一月十二'

    def test9(self):
        assert SolarDay(2023, 10, 27).get_term().__str__() == '霜降'

    def test10(self):
        assert SolarDay(2023, 10, 27).get_phenology_day().__str__() == '豺乃祭兽第4天'

    def test11(self):
        assert SolarDay(2023, 10, 27).get_phenology_day().get_phenology().get_three_phenology().__str__() == '初候'

    def test22(self):
        assert '甲辰' == SolarDay(2024, 2, 10).get_lunar_day().get_lunar_month().get_lunar_year().get_sixty_cycle().get_name()

    def test23(self):
        assert '癸卯' == SolarDay(2024, 2, 9).get_lunar_day().get_lunar_month().get_lunar_year().get_sixty_cycle().get_name()
